unit AgentClasses;

//  TCobBlock                         cob2 block
//   TAuthBlock                       cob2 author block
//   TFileBlock                       cob2 file dependancy block
//   TAgentBlock                      cob2 agent block
//  TCobFile                          cob2 file handling
//  TPRAYFile                         pray file handling
//  TPRAYBlock                        pray block
//   TNullPRAYBlock                   pray unknown block
//   TStandardPRAYBlock               pray normal block

interface

uses SysUtils, Dialogs, Windows, Classes, Forms, Graphics, ZeusClasses, unitErrorHandler;

type
  ECreaturesException = class(Exception);

  TCobBlock = class
  public
    ourlen: Cardinal;
    procedure ReadFromFile(F: TStream; len: Integer); virtual; abstract;
    procedure WriteToFile(F: TStream); virtual; abstract;
    function Caption: String; virtual; abstract;
  end;

  TAuthHeader = record
    Day, Month: Char;
    Year: Word;
    Version, Revision: Char;
  end;

  TAuthBlock = class(TCobBlock)
  public
    header: TAuthHeader;
    AuthorName, AuthorEMail, AuthorURL, AuthorComments: String;
    procedure ReadFromFile(F: TStream; len: Integer); override;
    procedure WriteToFile(F: TStream); override;
    function Caption: String; override;
  end;

  TFileBlock = class(TCobBlock)
  public
    FileType: Word;
    Reserved: Cardinal;
    FileSize: Cardinal;
    FileName: String;
    cdata: Pointer;
    procedure ReadFromFile(F: TStream; len: Integer); override;
    procedure WriteToFile(F: TStream); override;
    function Caption: String; override;
  end;

  TAgentHeader = record
    QuantityRemaining: Word;
    LastUsage: Cardinal;
    ReuseInterval: Cardinal;
    UseByDay: Char;
    UseByMonth: Char;
    UseByYear: Word;
    Reserved1: Cardinal;
    Reserved2: Cardinal;
    Reserved3: Cardinal;
  end;

  TAgentBlock = class(TCobBlock)
  public
    header: TAgentHeader;
    AgentName, AgentDescription, InstallScript, RemovalScript: String;
    NumberOfEvents: Word;
    EventScripts: array[0..200] of String; // max 200 event scripts
    NumberOfDeps: Word;
    FileDeps_type: array[0..200] of Word;
    FileDeps_name: array[0..200] of String;
    ThumbnailWidth, ThumbnailHeight: Word;
    Thumbnail: Pointer;
    procedure ReadFromFile(F: TStream; len: Integer); override;
    procedure WriteToFile(F: TStream); override;
    procedure ReadFromC1File(F: TStream);
    procedure WriteToC1File(F: TStream);
    function Caption: String; override;

    procedure WriteToStrings(strings: TStrings);
  end;

  // *** long-term fixme ***
  // you should split TCobFile into TCobFile abstract and TC2CobFile and TC1CobFile.
  TCobFile = class
  public
    currentfile: String;
    cobblocks: array[1..200] of TCobBlock;
    noblocks: Integer;
    c1format: Boolean;
    procedure Reset;
    procedure DeleteBlock(blockid: Integer);
    function GetAgentNo(no: Integer): Integer;
    function GetAgent(no: Integer): TAgentBlock;
    function GetAuthorInfo: TAuthBlock;
    procedure DeleteAuthorInfo;
    procedure ReadCobFile(F: TStream);
    procedure WriteCobFile(F: TStream);
  end;

  TCobHeader = record
    ChunkType: array[1..4] of Char;
    ChunkSize: Cardinal;
  end;

  TPRAYHeader = record
    PRAYtype: array[0..3] of Char;
    name: array[0..127] of Char;
    size: LongInt;
    usize: LongInt; // <-- this is only used at save/load time
    flags: LongInt; // <-- this is only used at save/load time
  end;

  TPRAYBlock = class
    header: TPRAYHeader;

    procedure Edit; virtual; abstract;
    procedure InitFromBuf(buf: Pointer); virtual; abstract;
    function InitToBuf: Pointer; virtual; abstract;
  end;

  TNullPRAYBlock = class(TPRAYBlock)
    data: Pointer;

    procedure Edit; override;
    procedure InitFromBuf(buf: Pointer); override;
    function InitToBuf: Pointer; override;
  end;

  TStandardPRAYBlock = class(TPRAYBlock)
    noints, nostrs: Integer;

    intnames: array[1..200] of string;
    strnames: array[1..200] of string;
    intvalues: array[1..200] of Integer;
    strvalues: array[1..200] of string;

    procedure Edit; override;
    procedure InitFromBuf(buf: Pointer); override;
    function InitToBuf: Pointer; override;
  end;

  TPRAYFile = class
    noblocks: Integer;
    blocks: array[0..200] of TPRAYBlock;

    procedure SaveToFile(F: TStream);
    procedure LoadFromFile(F: TStream);
  end;

function ReadStringFromFile(F: TStream): String;
procedure WriteStringToFile(F: TStream; data: String);

implementation

uses zlib, zutil, zinflate, zdeflate;

procedure TAgentBlock.WriteToStrings(strings: TStrings);

procedure process(b: string);
var
  z: string;
begin
  while not (b = '') do begin
    z := b;
    strings.Add(RemoveSpecial(b, ','));
  end;
  if not (LowerCase(z) = 'endm') then strings.Add('endm');
end;

var
  i: integer;
begin
  strings.Add('iscr');
  process(InstallScript);
  strings.Add('');
  strings.Add('rscr');
  process(RemovalScript);
  for i := 1 to NumberOfEvents do begin
    strings.Add('');
    process(EventScripts[i]);
  end;
end;

function ReadStringFromFile(F: TStream): String;
var
  i: Integer;
  cha: array[0..10000] of Char;
  x: Boolean;
begin
  i := 0;
  cha[i] := #0;
  x := True;
  while x do begin
    F.Read(cha[i], 1);
    if (cha[i] = #0) then x := False;
    i := i + 1;
  end;
  Result := cha;
end;

procedure WriteStringToFile(F: TStream; data: String);
var
  d: PChar;
begin
  GetMem(d, Length(data) + 1);
  // [cue mutters about StrPCopy not working on huge (> 255 chars) strings (in Delphi 2 anyway), and
  //  the fact that this problem is not documented ... and cue rememberance that this
  //  problem took TWO HOURS to find! <mutter mutter mutter>]
  StrLCopy(d, PChar(data), Length(data) + 1);
  F.Write(d^, Length(data) + 1);
  FreeMem(d);
end;

function TCobFile.GetAuthorInfo: TAuthBlock;
var
  i: Integer;
begin
  for i := 1 to noblocks do begin
    if cobblocks[i] is TAuthBlock then begin
      Result := TAuthBlock(cobblocks[i]);
      Exit;
    end;
  end;

  // okay, there is no author block. so we create one :)
  noblocks := noblocks + 1;
  cobblocks[noblocks] := TAuthBlock.Create;
  Result := TAuthBlock(cobblocks[noblocks]);
end;

procedure TCobFile.DeleteAuthorInfo;
var
  i: Integer;
begin
  for i := 1 to noblocks do begin
    if cobblocks[i] is TAuthBlock then begin
      DeleteBlock(i);
    end;
  end;
end;

procedure TCobFile.DeleteBlock(blockid: Integer);
begin
  noblocks := noblocks - 1;
  cobblocks[blockid].Free;
  cobblocks[blockid] := cobblocks[noblocks + 1];
end;

function TCobFile.GetAgentNo(no: Integer): Integer;
var
  i, i2: Integer;
begin
  i2 := -1;
  for i := 1 to noblocks do begin
    if cobblocks[i] is TAgentBlock then with cobblocks[i] as TAgentBlock do begin
      i2 := i2 + 1;
      if no = i2 then begin
        Result := i;
        Exit;
      end;
    end;
  end;

  Result := 0;
end;

function TCobFile.GetAgent(no: Integer): TAgentBlock;
var
  i, i2: Integer;
begin
  i2 := -1;
  for i := 1 to noblocks do begin
    if cobblocks[i] is TAgentBlock then with cobblocks[i] as TAgentBlock do begin
      i2 := i2 + 1;
      if no = i2 then begin
        Result := TAgentBlock(cobblocks[i]);
        Exit;
      end;
    end;
  end;

  Result := nil;
end;

procedure TCobFile.Reset;
begin
  noblocks := 0;
end;

procedure TCobFile.ReadCobFile(F: TStream);
var
  header: array[1..4] of Char;
  s: String;
  v: Integer;
  cobhead: TCobHeader;
begin
  Reset;
  F.Read(header, 4);
  s := header;
  if s = 'cob2' then begin
    c1format := False;
    while not (F.Read(cobhead, sizeof(cobhead)) = 0) do begin
      s := cobhead.ChunkType;
      if s = 'auth' then begin
        noblocks := noblocks + 1;
        cobblocks[noblocks] := TAuthBlock.Create;
        cobblocks[noblocks].ReadFromFile(F, cobhead.ChunkSize);
      end else if s = 'agnt' then  begin
        noblocks := noblocks + 1;
        cobblocks[noblocks] := TAgentBlock.Create;
        cobblocks[noblocks].ReadFromFile(F, cobhead.ChunkSize);
      end else if s = 'file' then  begin
        noblocks := noblocks + 1;
        cobblocks[noblocks] := TFileBlock.Create;
        cobblocks[noblocks].ReadFromFile(F, cobhead.ChunkSize);
      end else begin
        F.Seek(cobhead.ChunkSize, 1);
        ShowMessage('This file contains block type ' + s + '. This version of ZeusCOB doesn''t know about this type of block, so we''re ignoring it. This file *may* be corrupted.');
      end;
    end;
  end else begin
    c1format := True;
    F.Seek(0, 0);
    v := 0;
    F.Read(v, 2);
    if not (v = 1) then begin
      ErrorDialog(0, True); Exit;
    end;
    ShowMessage('You appear to be opening a Creatures 1 COB file. Please note that Creatures 1 support is currently in beta-test stage, and may not work perfectly.');
    noblocks := 1;
    cobblocks[1] := TAgentBlock.Create;
    with cobblocks[1] as TAgentBlock do ReadFromC1File(F);
  end;
end;

procedure TCobFile.WriteCobFile(F: TStream);
var
  i: Integer;
begin
  if c1format then begin
    ShowMessage('You appear to be saving a Creatures 1 COB file. Please note that Creatures 1 support is currently in beta-test stage, and may not work perfectly.');
    i := 1;
    F.Write(i, 2);
    with cobblocks[1] as TAgentBlock do WriteToC1File(F);
    Exit;
  end;

//  currentfile := filename;
  F.Write('cob2', 4);
  for i := 1 to noblocks do
  begin
    cobblocks[i].WriteToFile(F);
  end;
end;

procedure TAuthBlock.ReadFromFile(F: TStream; Len: Integer);
begin
  ourlen := Len;
  F.Read(header, sizeof(header));
  AuthorName := ReadStringFromFile(F);
  AuthorEMail := ReadStringFromFile(F);
  AuthorURL := ReadStringFromFile(F);
  AuthorComments := ReadStringFromFile(F);
end;

procedure TAuthBlock.WriteToFile(F: TStream);
var
  l: Cardinal;
begin
  F.Write('auth', 4);
  l := sizeof(header) + Length(AuthorName) + Length(AuthorEMail) + Length(AuthorURL) + Length(AuthorComments) + 4;
  F.Write(l, 4);
  F.Write(header, sizeof(header));
  WriteStringToFile(F, AuthorName);
  WriteStringToFile(F, AuthorEMail);
  WriteStringToFile(F, AuthorURL);
  WriteStringToFile(F, AuthorComments);
end;

function TAuthBlock.Caption: String;
begin
  Result := 'Author Information - [' + AuthorName + ']';
end;

procedure WriteWierdStringToFile(F: TStream; s: string);
var
  i: Integer;
  c: Char;
  x: PChar;
begin
  i := Length(s);
  if i > 254 then begin c := #255; F.Write(c, 1); F.Write(i, 2); end else F.Write(i, 1);
  x := PChar(s);
  F.Write(x^, i);
end;

function ReadWierdStringFromFile(F: TStream): string;
var
  c: Char;
  x, i: Integer;
begin
  F.Read(c, 1);
  x := 0;
  if c = #255 then F.Read(x, 2) else x := Integer(c);

  Result := '';
  for i := 1 to x do begin
    F.Read(c, 1);
    Result := Result + c;
  end;
end;

procedure TAgentBlock.WriteToC1File(F: TStream);
var
  i: Integer;
begin
  F.Write(header.QuantityRemaining, 2);
  F.Write(header.UseByMonth, 4);
  F.Write(header.UseByDay, 4);
  F.Write(header.UseByYear, 4);

  F.Write(NumberOfEvents, 2);
  i := 1; // 1 import (injection script)
  F.Write(i, 2);

  i := 0;
  F.Write(i, 4);

  for i := 1 to NumberOfEvents do begin
    WriteWierdStringToFile(F, EventScripts[i]);
  end;
  WriteWierdStringToFile(F, InstallScript);
  // *** fixme *** removal script should be written to the RCB

  // write the thumbnail :)
  ShowMessage('Warning: Creatures 1 thumbnails are not supported at this time. The thumbnail will be blanked out.');
  ThumbnailWidth := 0;
  ThumbnailHeight := 0;

  F.Write(ThumbnailWidth, 4);
  F.Write(ThumbnailHeight, 4);
  i := 0;
  F.Write(i, 2);

  WriteWierdStringToFile(F, AgentName);
end;

procedure TAgentBlock.ReadFromC1File(F: TStream);
var
  imp, i: Integer;
begin
  header.QuantityRemaining := 0;
  F.Read(header.QuantityRemaining, 2);
  header.LastUsage := 0;
  header.ReuseInterval := 0;
  F.Read(header.UseByMonth, 4);
  F.Read(header.UseByDay, 4);
  F.Read(header.UseByYear, 4);
  F.Read(NumberOfEvents, 2);
  NumberOfDeps := 0; // deps not allowed :)
  imp := 0;
  F.Read(imp, 2);
  if imp > 1 then ErrorDialog(1, False);
  F.Read(i, 4);
  for i := 1 to NumberOfEvents do begin
    EventScripts[i] := ReadWierdStringFromFile(F);
  end;
  for i := 1 to imp do begin
    if i = 1 then InstallScript := ReadWierdStringFromFile(F);
  end;
  RemovalScript := ''; // *** fixme *** : script should load from the remover file (RCB)

  // read the thumbnail :)
  ShowMessage('Warning: Creatures 1 thumbnails are not supported at this time. The thumbnail will be blanked out.');
  F.Read(ThumbnailWidth, 4);
  F.Read(ThumbnailHeight, 4);
  F.Read(i, 2);
  F.Seek(ThumbnailWidth * ThumbnailHeight, 1);
  ThumbnailWidth := 0;
  ThumbnailHeight := 0;
  AgentName := ReadWierdStringFromFile(F);
  AgentDescription := '';
end;

procedure TAgentBlock.ReadFromFile(F: TStream; Len: Integer);
var
  i: Integer;
begin
  ourlen := Len;
  F.Read(header.QuantityRemaining, 2);
  F.Read(header.LastUsage, 4);
  F.Read(header.ReuseInterval, 4);
  F.Read(header.UseByDay, 1);
  F.Read(header.UseByMonth, 1);
  F.Read(header.UseByYear, 2);
  F.Read(header.Reserved1, 4);
  F.Read(header.Reserved2, 4);
  F.Read(header.Reserved3, 4);
  AgentName := ReadStringFromFile(F);
  AgentDescription := ReadStringFromFile(F);
  InstallScript := ReadStringFromFile(F);
  RemovalScript := ReadStringFromFile(F);
  F.Read(NumberOfEvents, sizeof(NumberOfEvents));
  for i := 1 to NumberOfEvents do
  begin
    EventScripts[i] := ReadStringFromFile(F);
  end;
  F.Read(NumberOfDeps, sizeof(NumberOfDeps));
  for i := 1 to NumberOfDeps do
  begin
    F.Read(FileDeps_type[i], sizeof(FileDeps_type[i]));
    FileDeps_name[i] := ReadStringFromFile(F);
  end;
  F.Read(ThumbnailWidth, sizeof(ThumbnailWidth));
  F.Read(ThumbnailHeight, sizeof(ThumbnailHeight));
  GetMem(Thumbnail, ThumbnailWidth * ThumbnailHeight * 2);
  F.Read(Thumbnail^, ThumbnailWidth * ThumbnailHeight * 2);
end;

procedure TAgentBlock.WriteToFile(F: TStream);
var
  i: Integer;
  l: Cardinal;
begin
  F.Write('agnt', 4);
//  FileWrite(F, ourlen, sizeof(ourlen));
  l := 26 + Length(AgentName) + Length(AgentDescription) + Length(InstallScript) + Length(RemovalScript) + 4;
  l := l + sizeof(NumberOfEvents) + sizeof(NumberOfDeps) + sizeof(ThumbNailWidth) + sizeof(ThumbnailHeight);
  l := l + (ThumbnailWidth * ThumbnailHeight * 2);
  for i := 1 to NumberOfEvents do
  begin
    l := l + Length(EventScripts[i]) + 1;
  end;
  for i := 1 to NumberOfDeps do
  begin
    l := l + sizeof(FileDeps_type[i]) + Length(FileDeps_name[i]) + 1;
  end;
  F.Write(l, sizeof(l));
  F.Write(header.QuantityRemaining, 2);
  F.Write(header.LastUsage, 4);
  F.Write(header.ReuseInterval, 4);
  F.Write(header.UseByDay, 1);
  F.Write(header.UseByMonth, 1);
  F.Write(header.UseByYear, 2);
  F.Write(header.Reserved1, 4);
  F.Write(header.Reserved2, 4);
  F.Write(header.Reserved3, 4);
  WriteStringToFile(F, AgentName);
  WriteStringToFile(F, AgentDescription);
  WriteStringToFile(F, InstallScript);
  WriteStringToFile(F, RemovalScript);
  F.Write(NumberOfEvents, sizeof(NumberOfEvents));
  for i := 1 to NumberOfEvents do
  begin
    WriteStringToFile(F, eventscripts[i]);
  end;
  F.Write(NumberOfDeps, sizeof(NumberOfDeps));
  for i := 1 to NumberOfDeps do
  begin
    F.Write(FileDeps_type[i], sizeof(FileDeps_type[i]));
    WriteStringToFile(F, FileDeps_name[i]);
  end;
  F.Write(ThumbnailWidth, sizeof(ThumbnailWidth));
  F.Write(ThumbnailHeight, sizeof(ThumbnailHeight));
  F.Write(Thumbnail^, ThumbnailWidth * ThumbnailHeight * 2);
end;

function TAgentBlock.Caption: String;
begin
  Result := 'Agent - [' + AgentName + ']';
end;

procedure TFileBlock.ReadFromFile(F: TStream; Len: Integer);
begin
  ourlen := Len;
  F.Read(FileType, 2);
  F.Read(Reserved, 4);
  F.Read(FileSize, 4);
  FileName := ReadStringFromFile(F);
  GetMem(cdata, FileSize);
  F.Read(cdata^, FileSize);
end;

procedure TFileBlock.WriteToFile(F: TStream);
var
  l: Cardinal;
begin
  F.Write('file', 4);
  l := 10 + Length(FileName) + 1 + FileSize;
  F.Write(l, 4);
  F.Write(FileType, 2);
  F.Write(Reserved, 4);
  F.Write(FileSize, 4);
  WriteStringToFile(F, FileName);
  F.Write(cdata^, FileSize);
end;

function TFileBlock.Caption: String;
begin
  Result := 'File Dependancy - [' + FileName + ']';
end;

function compress (dest : pBytef; var destLen : uLong; source : pBytef; sourceLen : uLong; level : int) : int;
var
  stream : z_stream;
  err : int;
begin
  stream.next_in := pBytef(source);
  stream.avail_in := uInt(sourceLen);

  stream.next_out := dest;
  stream.avail_out := uInt(destLen);
  if (uLong(stream.avail_out) <> destLen) then begin
    compress := Z_BUF_ERROR;
    exit;
  end;

  stream.zalloc := NIL;       { alloc_func(0); }
  stream.zfree := NIL;        { free_func(0); }
  stream.opaque := NIL;       { voidpf(0); }

  err := deflateInit(stream, level);
  if (err <> Z_OK) then begin
    compress := err;
    exit;
  end;

  err := deflate(stream, Z_FINISH);
  if (err <> Z_STREAM_END) then begin
    deflateEnd(stream);
    if err = Z_OK then
      compress := Z_BUF_ERROR
    else
      compress := err;
    exit;
  end;
  destLen := stream.total_out;

  err := deflateEnd(stream);
  compress := err;
end;

function uncompress(dest: pBytef; destLen: uLong; source: pBytef; sourceLen: uLong): int;
var
  stream : z_stream;
  err : int;
begin
  stream.next_in := source;
  stream.avail_in := uInt(sourceLen);
  stream.next_out := dest;
  stream.avail_out := uInt(destLen);

  stream.zalloc := nil;
  stream.zfree := nil;

  err := inflateInit(stream);
  if err <> Z_OK then begin
    ErrorDialog(2, True);
    Result := err;
    exit;
  end;

  err := inflate(stream, Z_FINISH);
  if err <> Z_STREAM_END then begin
    ErrorDialog(3, True);
    inflateEnd(stream);
    if err = Z_OK then
      Result := Z_BUF_ERROR
    else
      Result := err;
    exit;
  end;
//  destLen := stream.total_out;

  err := inflateEnd(stream);
  Result := err;
end;

function NewPRAYBlock(hdr: TPRAYHeader; buf: Pointer): TPRAYBlock;
begin
  if hdr.PRAYType = 'AGNT' then
    Result := TStandardPRAYBlock.Create
  else if hdr.PRAYType = 'EGGS' then
    Result := TStandardPRAYBlock.Create
  else begin
    Result := TNullPRAYBlock.Create;
  end;

  Result.InitFromBuf(buf);
  Result.header := hdr;
end;

procedure TStandardPRAYBlock.InitFromBuf(buf: Pointer);
var
  ptr: Pointer;
  o, j, len: Integer;
  i: array[0..50000] of Char;
begin
  ptr := buf;
  o := 0;

  Move(Pointer(Integer(ptr) + o)^, noints, 4);
  o := o + 4;

  for j := 1 to noints do begin
    Move(Pointer(Integer(ptr) + o)^, len, 4);
    o := o + 4;

    Move(Pointer(Integer(ptr) + o)^, i, len);
    o := o + len;
    i[len] := #0;

    intnames[j] := i;

    Move(Pointer(Integer(ptr) + o)^, len, 4);
    o := o + 4;

    intvalues[j] := len;
  end;

  Move(Pointer(Integer(ptr) + o)^, nostrs, 4);
  o := o + 4;

  for j := 1 to nostrs do begin
    Move(Pointer(Integer(ptr) + o)^, len, 4);
    o := o + 4;

    Move(Pointer(Integer(ptr) + o)^, i, len);
    o := o + len;
    i[len] := #0;

    strnames[j] := i;

    Move(Pointer(Integer(ptr) + o)^, len, 4);
    o := o + 4;

    Move(Pointer(Integer(ptr) + o)^, i, len);
    o := o + len;
    i[len] := #0;

    strvalues[j] := i;
  end;
end;

function TStandardPRAYBlock.InitToBuf: Pointer;
var
  bptr: Pointer;
  i, j, o: Integer;
begin
  // step one: work out our size
  header.size := 8; // nostrs and noints
  for j := 1 to nostrs do begin;
    header.size := header.size + 8 + Length(strnames[j]) + Length(strvalues[j]);
  end;
  for j := 1 to noints do begin;
    header.size := header.size + 8 + Length(intnames[j]);
  end;

  GetMem(bptr, header.size);

  // step two: copy everything in
  o := 0;

  Move(noints, Pointer(Integer(bptr) + o)^, 4);
  o := o + 4;
  for j := 1 to noints do begin
    i := Length(intnames[j]);
    Move(i, Pointer(Integer(bptr) + o)^, 4);
    o := o + 4;
    Move(PChar(intnames[j])^, Pointer(Integer(bptr) + o)^, i);
    o := o + i;

    Move(intvalues[j], Pointer(Integer(bptr) + o)^, 4);
    o := o + 4;
  end;

  Move(nostrs, Pointer(Integer(bptr) + o)^, 4);
  o := o + 4;
  for j := 1 to nostrs do begin
    i := Length(strnames[j]);
    Move(i, Pointer(Integer(bptr) + o)^, 4);
    o := o + 4;
    Move(PChar(strnames[j])^, Pointer(Integer(bptr) + o)^, i);
    o := o + i;

    i := Length(strvalues[j]);
    Move(i, Pointer(Integer(bptr) + o)^, 4);
    o := o + 4;
    Move(PChar(strvalues[j])^, Pointer(Integer(bptr) + o)^, i);
    o := o + i;
  end;

  Result := bptr;
end;

procedure TStandardPRAYBlock.Edit;
begin
end;

procedure TNullPRAYBlock.InitFromBuf(buf: Pointer);
begin
  data := buf;
end;

function TNullPRAYBlock.InitToBuf: Pointer;
begin
  Result := data;
end;

procedure TNullPRAYBlock.Edit;
begin
  // fixme: should it be name or type?
  ShowMessage('Zeus doesn''t know how to edit PRAY blocks of type ' + header.name + '!');
end;

procedure TPRAYFile.SaveToFile(F: TStream);
var
  majic: array[0..3] of Char;
  i: Integer;
  src, buffer: Pointer;
  c: uLong;
begin
  majic := 'PRAY';

  // write da header
  F.Write(majic, sizeof(majic));

  for i := 0 to noblocks do begin
    buffer := blocks[i].InitToBuf;
    blocks[i].header.usize := blocks[i].header.size;
    if (blocks[i].header.flags and 1) = 1 then begin // compression
      // flip the buffer to our src buffer
      src := buffer;
      // allocate a buffer
      c := blocks[i].header.usize + 1000;
      GetMem(buffer, c);

      // compress from source to buffer
      if compress(buffer, c, src, blocks[i].header.usize, 9) <> Z_OK then // always use max compression
        ErrorDialog(4, True);

      // set the header size up prply
      blocks[i].header.size := c;

      // free the source buffer
      FreeMem(src);
    end;
    F.Write(blocks[i].header, sizeof(blocks[i].header));
    F.Write(buffer^, blocks[i].header.size);
    FreeMem(buffer);
  end;
end;

procedure TPRAYFile.LoadFromFile(F: TStream);
var
  hdr: TPRAYHeader;
  majic: array[0..3] of Char;
  src, buffer: Pointer;
begin
  noblocks := -1;

    // read the header
    F.Read(majic, sizeof(majic));
    if majic <> 'PRAY' then begin ShowMessage('This file is not a PRAY file.'); Exit; end;

    while F.Read(hdr, sizeof(hdr)) = sizeof(hdr) do begin
      GetMem(src, hdr.size);
      if src = nil then begin ErrorDialog(6, True); Continue; end;
      try
        F.Read(src^, hdr.size);
        if (hdr.flags and 1) = 1 then begin
          // allocate buffer, uncompress
          GetMem(buffer, hdr.usize);
          if buffer = nil then begin ErrorDialog(7, True); Continue; end;
          uncompress(buffer, hdr.usize, src, hdr.size);

          noblocks := noblocks + 1;
          blocks[noblocks] := NewPRAYBlock(hdr, buffer);
        end else begin
          // buffer is already uncompressed

          noblocks := noblocks + 1;
          blocks[noblocks] := NewPRAYBlock(hdr, src);
        end;
      finally
        if (hdr.flags and 1) = 1 then
          FreeMem(src);
      end;
    end;
end;

end.
