unit ZeusClasses;

interface

uses Classes, SysUtils, CreatureLabsCommLink, Forms, Windows;

type
  TCAOSParseEvent = procedure(data: string; iscr, rscr: boolean);

  TZeusSerializableObject = class
  public
    procedure SerializeFromStream(stream: TStream); virtual; abstract;
    procedure SerializeToStream(stream: TStream); virtual; abstract;
  end;

  TZeusFile = class(TZeusSerializableObject)
  public
    // fname is only valid during serialization
    fname: string;

    procedure SerializeFromFile(filename: string);
    procedure SerializeToFile(filename: string);
  end;

var
  CommLink: TCreatureLabsCommLink;

function ExtractNoExt(b: string): string;
function RemoveSpecial(var fromm: string; c: Char): String;
procedure SyntaxDataUpdate;

implementation

uses unitSyntaxInformation, unitCAOSReference{, GeneClasses};

procedure SyntaxDataUpdate;
begin
  syntaxdata.Free;

  syntaxdata := TSyntaxData.Create;
  try
    syntaxdata.ReadFromFile(ExtractFileDir(Application.EXEName) + '\' + CommLink.GameName + '\syntaxdata.txt');
  except
    on Exception do Application.MessageBox(PChar('No syntax data is available for game ' + CommLink.GameName + ', so syntax checking will not be operational.'), 'Zeus', MB_OK + MB_ICONWARNING);
  end;
end;

function RemoveSpecial(var fromm: string; c: Char): String;
begin
  if pos(c, fromm) = 0 then
  begin
    Result := fromm;
    fromm := '';
    Exit;
  end;
  Result := copy(fromm, 1, pos(c, fromm) - 1);
  fromm := copy(fromm, pos(c, fromm) + 1, length(fromm) - pos(c, fromm));
end;

function ExtractNoExt(b: string): string;
var
  p: string;
begin
  p := ExtractFileExt(b);
  Result := Copy(ExtractFileName(b), 1, Length(ExtractFileName(b)) - Length(p));
end;

procedure TZeusFile.SerializeFromFile(filename: string);
var
  s: TStream;
begin
  fname := filename;

  s := TFileStream.Create(filename, fmOpenRead);
  try
    SerializeFromStream(s);
  finally
    s.Free;
    fname := '';
  end;
end;

procedure TZeusFile.SerializeToFile(filename: string);
var
  s: TStream;
begin
  fname := filename;

  s := TFileStream.Create(filename, fmCreate);
  try
    SerializeToStream(s);
  finally
    s.Free;
    fname := '';
  end;
end;

initialization
  syntaxdata := TSyntaxData.Create;
  CommLink := TCreatureLabsCommLink.Create(nil);
finalization
  CommLink.Free; syntaxdata.Free;
end.
