unit BLKClasses;

interface

uses Windows, Graphics, SysUtils, Dialogs, unitErrorHandler;

type
  TBLKFile = class
  public
    hasdata: Boolean;
    data: array[1..1000] of PChar;
    nosprites: Integer;
    bheight, bwidth: Integer;
    widths, heights: array[1..1000] of Integer;
    is_565: boolean;
    TotalWidth, TotalHeight: Integer;

    procedure Open(fname: string);
    procedure Paint(canv: TCanvas);

    constructor Create;
    destructor Destroy; override;
  end;

  TBitmapInfo16 = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array[0..2] of Cardinal; // 3 16-bit bit-masks... R, G and B :)
  end;

  TS16Frame = class
  public
    width, height: Word;
    data: PChar;
    is_565: Boolean;

    procedure ConvertTo555;
    procedure ConvertTo565;

    // note: the canvas passed to both functions should be the FORM canvas!
    procedure LoadFromBMP(bitmap: TBitmap; canv: TCanvas);
  end;

  // this is a cut-down version of the Zeus TCreaturesImage
  TS16File = class
    noframes: Integer;
    frames: array[1..2000] of TS16Frame;

    is_565: Boolean;

    procedure ConvertTo555;
    procedure ConvertTo565;

    procedure LoadFromS16(filename: string);
    procedure FreeData;
  end;

function CreateDIBitmap16(DC: HDC; var InfoHeader: TBitmapInfoHeader; dwUsage: DWORD; InitBits: PChar; var InitInfo: TBitmapInfo16; wUsage: UINT): HBITMAP; stdcall; external 'gdi32.dll' name 'CreateDIBitmap';
function GetDIBits16(DC: HDC; Bitmap: HBitmap; StartScan, NumScans: UINT; Bits: Pointer; var BitInfo: TBitmapInfo16; Usage: UINT): Integer; stdcall; external 'gdi32.dll' name 'GetDIBits';
procedure DrawFrame(canvas: TCanvas; is_565: boolean; width, height: LongInt; data: PChar; x, y: Integer);

procedure WriteS16AsBLK(fname: string; src: TS16File; bwidth, bheight: Integer);
function TranslateData(data: PChar; width, height: integer): TS16File;
function TranslateBMPToData(is_565: Boolean; bit: TBitmap): PChar;

implementation

procedure TS16File.FreeData;
var
  i: Integer;
begin
  for i := 1 to noframes do begin
    if frames[i].data <> nil then
      FreeMem(frames[i].data);
  end;
end;

procedure TS16Frame.LoadFromBMP(bitmap: TBitmap; canv: TCanvas);
var
  bitinfo: TBitmapInfo16;
  buffersize{, i}: Integer;
//  x: PWord;
//  y: Word;
//  r, g, b: Word;
begin
  width := bitmap.width;
  height := bitmap.height;

  bitinfo.bmiHeader.biSize := sizeof(bitinfo.bmiHeader);
  bitinfo.bmiHeader.biWidth := width;
  bitinfo.bmiHeader.biHeight := -height;
  bitinfo.bmiHeader.biPlanes := 1;
  bitinfo.bmiHeader.biClrUsed := 0;
  bitinfo.bmiHeader.biClrImportant := 0;
  bitinfo.bmiHeader.biBitCount := 16;
  bitinfo.bmiHeader.biCompression := BI_BITFIELDS;
  bitinfo.bmiHeader.biSizeImage := width * height * 2;
  buffersize := width * height * 2;
  GetMem(data, buffersize);
  GetDIBits16(canv.Handle, bitmap.Handle, 0, height, data, bitinfo, DIB_RGB_COLORS);
end;

procedure TS16File.ConvertTo555;
var
  i: Integer;
begin
  for i := 1 to noframes do begin
    frames[i].ConvertTo555;
  end;

  is_565 := false;
end;

procedure TS16File.ConvertTo565;
var
  i: Integer;
begin
  for i := 1 to noframes do begin
    frames[i].ConvertTo565;
  end;

  is_565 := true;
end;

procedure TS16Frame.ConvertTo555;
var
  i: Integer;
  x: PWord;
  r, g, b, y: Word;
begin
  if not is_565 then Exit;

  for i := 1 to (width * height) do begin
    x := PWord(Integer(data) + ((i - 1) * 2));
    y := x^;

    r := (y and $f800) shr 11;
    g := (y and $7e0) shr 6;
    b := y and $1f;

    g := g shl 5;
    r := r shl 10;
    y := b + (g and $3e0) + (r and $7c00);

    x^ := y;
  end;
end;

procedure TS16Frame.ConvertTo565;
var
  i: Integer;
  x: PWord;
  r, g, b, y: Word;
begin
  if is_565 then Exit;

  for i := 1 to (width * height) do begin
    x := PWord(Integer(data) + ((i - 1) * 2));
    y := x^;

    r := (y and $7c00) shr 10;
    g := (y and $3e0) shr 5;
    b := y and $1f;

    g := g shl 6;
    r := r shl 11;
    y := b + (g and $7e0) + (r and $F800);

    x^ := y;
  end;
end;

procedure TS16File.LoadFromS16(filename: string);
var
  F, i: Integer;
  flags: Cardinal;
  spritecount: Word;
  offsets: array[0..2000] of Cardinal;
  buffersize: Integer;
  width, height: Word;
begin
  F := FileOpen(filename, fmOpenRead);
  FileRead(F, flags, 4);
  FileRead(F, spritecount, 2);
  noframes := spritecount;

  is_565 := flags = 1;

  for i := 1 to spritecount do begin
    FileRead(F, offsets[i], 4);
    frames[i] := TS16Frame.Create;
    frames[i].is_565 := flags = 1;
    FileRead(F, width, 2);
    frames[i].width := width;
    FileRead(F, height, 2);
    frames[i].height := height;
  end;

  for i := 1 to spritecount do begin
    FileSeek(F, offsets[i], 0);
    buffersize := frames[i].width * frames[i].height;
    GetMem(frames[i].data, buffersize * 2);
    if (frames[i].data = nil) then begin ErrorDialog(12, True); Exit; end;
    FileRead(F, frames[i].data^, buffersize * 2);
  end;

  FileClose(F);
end;

procedure WriteS16AsBLK(fname: string; src: TS16File; bwidth, bheight: Integer);
var
  F, i, offset: Integer;
  flags: Integer;
begin
  // fixme: throw an exception if the blocks aren't 128x128

  F := FileCreate(fname);
  if src.is_565 then flags := 1 else flags := 0;
  FileWrite(F, flags, 4);
  FileWrite(F, bwidth, 2);
  FileWrite(F, bheight, 2);
  FileWrite(F, src.noframes, 2);

  offset := 6 + (src.noframes * 8); // for some reason, the width/height are removed from the offset??? wierd CLabs programmers... :)

  for i := 1 to src.noframes do begin
    FileWrite(F, offset, 4);
    offset := offset + (src.frames[i].width * src.frames[i].height * 2);
    FileWrite(F, src.frames[i].width, 2);
    FileWrite(F, src.frames[i].height, 2);
  end;

  for i := 1 to src.noframes do begin
    FileWrite(F, src.frames[i].data^, (src.frames[i].width * src.frames[i].height * 2));
  end;

  FileClose(F);
end;

function TranslateData(data: PChar; width, height: integer): TS16File;
var
  x, y: integer;
  r: TS16File;
  frame: TS16Frame;
  y2: integer;
begin
  r := TS16File.Create;

  // the hard bit - fortunately the input data is guaranteed to be 128x128 compatible (for now) =)
  // the way to do it? grab the data scanline-by-scanline

  for x := 0 to width - 1 do begin
    for y := 0 to height - 1 do begin
      Inc(r.noframes);
      frame := TS16Frame.Create;
      r.frames[r.noframes] := frame;
      frame.width := 128;
      frame.height := 128;
      GetMem(frame.data, 128 * 128 * 2);
      for y2 := 0 to 127 do begin
        Move((data
             + (width * 128 * ((y * 128) + y2) * 2)
             + (128 * x * 2))^, // this is correct =D
             (frame.data + (y2 * 128 * 2))^,
             128 * 2);
      end;
    end;
  end;

  Result := r;
end;

function TranslateBMPToData(is_565: Boolean; bit: TBitmap): PChar;
begin
  // unimplemented
  Result := nil;
end;

constructor TBLKFile.Create;
begin
  nosprites := 0;
  hasdata := false;
end;

destructor TBLKFile.Destroy;
var
  i: Integer;
begin
  if not hasdata then Exit;

  for i := 1 to nosprites do begin
    FreeMem(data[i]);
  end;
end;

procedure TBLKFile.Paint(canv: TCanvas);
var
  i: Integer;
  x, y: Integer;
begin
  if not hasdata then Exit;

  SelectClipRgn(canv.Handle, 0);
  x := 0; y := 0;
  for i := 1 to nosprites do begin
    DrawFrame(canv, is_565, widths[i], heights[i], data[i], x * 128, y * 128);
    Inc(y);
    if y = bheight then begin y := 0; Inc(x) end;
  end;
end;

procedure TBLKFile.Open(fname: string);
var
  F: Integer;
  flags: Integer;
  i: integer;
  offsets: array[1..1000] of Integer;
begin
  F := FileOpen(fname, fmOpenRead + fmShareDenyNone);
  if F = -1 then begin if FileExists(fname) then ErrorDialog(8, True) else ErrorDialog(9, True); hasdata := false; Exit; end;
  FileRead(F, flags, 4);
  is_565 := flags = 1;
  bwidth := 0;
  FileRead(F, bwidth, 2);
  bheight := 0;
  FileRead(F, bheight, 2);
  TotalWidth := bwidth * 128;
  TotalHeight := bheight * 128;
  nosprites := 0;
  FileRead(F, nosprites, 2);

  for i := 1 to nosprites do begin
    FileRead(F, offsets[i], 4);
    offsets[i] := offsets[i] + 4;
    widths[i] := 0;
    FileRead(F, widths[i], 2);
    heights[i] := 0;
    FileRead(F, heights[i], 2);
  end;

  for i := 1 to nosprites do begin
    FileSeek(F, offsets[i], 0);
    GetMem(data[i], widths[i] * heights[i] * 2);
    FileRead(F, data[i]^, widths[i] * heights[i] * 2);
  end;

//  Form1.StatusBar1.SimpleText := 'Sprites :' + IntToStr(nosprites) + ', Width: 128 * ' + IntToStr(bwidth) + ' (' + IntToStr(bwidth * 128) + '), Height: 128 * '  + IntToStr(bheight) + ' (' + IntToStr(bheight * 128) + ')';

  FileClose(F);

  hasdata := true;
end;

procedure DrawFrame(canvas: TCanvas; is_565: boolean; width, height: LongInt; data: PChar; x, y: Integer);
var
  mybit: TBitmap;
  bitinfo: TBitmapInfo16;
begin
  mybit := TBitmap.Create;

  bitinfo.bmiHeader.biSize := sizeof(bitinfo.bmiHeader);
  bitinfo.bmiHeader.biWidth := width;
  bitinfo.bmiHeader.biHeight := -height;
  bitinfo.bmiHeader.biPlanes := 1;
  bitinfo.bmiHeader.biClrUsed := 0;
  bitinfo.bmiHeader.biClrImportant := 0;
  bitinfo.bmiHeader.biBitCount := 16;
  bitinfo.bmiHeader.biCompression := BI_BITFIELDS;
  bitinfo.bmiHeader.biSizeImage := width * height * 2;
  if not is_565 then begin // if its 555 ... use the 555 bit masks
    bitinfo.bmiColors[0] := $7C00;
    bitinfo.bmiColors[1] := $03E0;
    bitinfo.bmiColors[2] := $001F;
  end else begin // if its 565 ... guess :)
    bitinfo.bmiColors[0] := $F800;
    bitinfo.bmiColors[1] := $07E0;
    bitinfo.bmiColors[2] := $001F;
  end;
  mybit.Handle := CreateDIBitmap16(Canvas.Handle, bitinfo.bmiHeader, CBM_INIT, PChar(data), bitinfo, DIB_RGB_COLORS);

  if Integer(mybit.Handle) <> 0 then begin
    Canvas.Draw(x, y, mybit);
    mybit.Free;
  end;
end;

end.
