unit unitAgentForm;

// TAgentForm : Agent Editor

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, unitChildForm, Menus, StdCtrls, CAOSSyntaxMemo, ComCtrls, Registry, ExtCtrls, unitErrorHandler;

type
  TAgentForm = class(TMDIChildForm)
    SyntaxMemo: TSyntaxMemo; MainMenu1: TMainMenu; ListBox1: TListBox;
    Edit1: TMenuItem; View1: TMenuItem; ColorCoding1: TMenuItem; CAOSRefSync1: TMenuItem;
    ShowSyntaxArea1: TMenuItem; SyntaxChecking1: TMenuItem; Undo1: TMenuItem; N1: TMenuItem;
    Cut1: TMenuItem; Copy1: TMenuItem; Paste1: TMenuItem; Delete1: TMenuItem;
    SelectAll1: TMenuItem; Tools1: TMenuItem; InjectintoCreatures1: TMenuItem; InjectRemovalScript1: TMenuItem;
    N2: TMenuItem; CompiletoPRAY1: TMenuItem; Indent1: TMenuItem; N3: TMenuItem;
    N4: TMenuItem; Find1: TMenuItem; Replace1: TMenuItem; SyntaxCheck1: TMenuItem;
    AutoIndent1: TMenuItem; InjectEventScripts1: TMenuItem;
    SaveDialog1: TSaveDialog; FindDialog1: TFindDialog; ReplaceDialog1: TReplaceDialog;

    procedure FormCreate(Sender: TObject);
    procedure Cut1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure SyntaxMemoFormatLine(I: Integer; const M: string; var F: TFormatLineArray);
    procedure SyntaxMemoChange(Sender: TObject);
    procedure SyntaxMemoKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Indent1Click(Sender: TObject);
    procedure InjectintoCreatures1Click(Sender: TObject);
    procedure View1Click(Sender: TObject);
    procedure ColorCoding1Click(Sender: TObject);
    procedure CAOSRefSync1Click(Sender: TObject);
    procedure SyntaxChecking1Click(Sender: TObject);
    procedure Undo1Click(Sender: TObject);
    procedure Delete1Click(Sender: TObject);
    procedure SelectAll1Click(Sender: TObject);
    procedure CompiletoPRAY1Click(Sender: TObject);
    procedure InjectRemovalScript1Click(Sender: TObject);
    procedure Find1Click(Sender: TObject);
    procedure Replace1Click(Sender: TObject);
    procedure FindDialog1Find(Sender: TObject);
    procedure ReplaceDialog1Replace(Sender: TObject);
    procedure SyntaxCheck1Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ShowSyntaxArea1Click(Sender: TObject);
    procedure ListBox1Click(Sender: TObject);
    procedure SyntaxMemoEnter(Sender: TObject);
    procedure SyntaxMemoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure AutoIndent1Click(Sender: TObject);
    procedure InjectEventScripts1Click(Sender: TObject);
  public
    errorinfo: array[1..64000] of string; // fixme this should be dynamic or using the Data field in memo

    function IsFileBased: Boolean; override;
    procedure New; override;
    procedure Open(fname: string); override;
    procedure Save; override;

    procedure LineChangeSync;
  end;

var
  blue, err, black, red, purple, green, bold: TFont;
  AgentForm: TAgentForm;
  globalfontname: string; globalfontsize: Integer;

  SyntaxAreaVisible, ColorCode, SyntaxCheck, CAOSRefSync, autoindent: Boolean;

procedure SplitScriptStrings(src, install, remove, events: TStrings);

implementation

uses unitCAOSReference, unitPRAYWizard, AgentClasses, ZeusClasses, Unit1, unitTokeniser, unitSyntaxInformation, unitSyntaxCheck, unitCAOSCalculator;

{$R *.DFM}

function TAgentForm.IsFileBased: Boolean;
begin
  Result := True;
end;

procedure TAgentForm.New;
begin
  inherited;
  SyntaxMemo.Lines.Clear;
end;

procedure TAgentForm.Open(fname: string);
begin
  inherited;
  SyntaxMemo.Lines.LoadFromFile(filename);
  modified := false;
end;

procedure TAgentForm.Save;
begin
  if filename = '' then begin SaveAs; Exit; end;
  modified := false;
  SyntaxMemo.Lines.SaveToFile(filename);
end;

procedure TAgentForm.FormCreate(Sender: TObject);
begin
  inherited;

  SyntaxMemo.Font.Name := globalfontname;
  SyntaxMemo.Font.Size := globalfontsize;

  ListBox1.Visible := SyntaxAreaVisible;
end;

procedure TAgentForm.Cut1Click(Sender: TObject);
begin
  SyntaxMemo.CutToClipboard;
end;

procedure TAgentForm.Copy1Click(Sender: TObject);
begin
  SyntaxMemo.CopyToClipboard;
end;

procedure TAgentForm.Paste1Click(Sender: TObject);
begin
  SyntaxMemo.PasteFromClipboard;
end;

// format a line of CAOS code
procedure TAgentForm.SyntaxMemoFormatLine(I: Integer; const M: string; var F: TFormatLineArray);
var
  t: TTokenSet;
  j: Integer; // where we are in the string
  k: Integer; // where we are in the FormatLineArray
  p: Integer; // which token we're at
  s: string;
  err: TCAOSErrorData;
  line: Integer;
begin
  line := SendMessage(SyntaxMemo.Handle, EM_LINEFROMCHAR, SyntaxMemo.SelStart, 0);

  if not ColorCode then Exit;

  errorinfo[i] := '';

  // quick checks for speed
  if Length(Trim(m)) = 0 then Exit;
  if Trim(m)[1] = '*' then begin
    F[0].Start := 0; // make sure it aligns
    F[0].Length := Length(m);
    F[0].Font := green;
    Exit;
  end;

  s := LowerCase(m); // always work using lower-case versions

  t := TTokenSet.Create;
  try try
    try
      t.SlurpTokenSet(Trim(s));
    except
      // fixme: we crash if we continue with a closed string =(
      on g: ECAOSStringClosed do begin
        F[0].Start := 0;
        F[0].Length := Length(s);
        F[0].Font := unitAgentForm.err;
        errorinfo[i] := g.message;
        Exit;
      end;

      on e: Exception do begin
        // MAJOR syntax error on this line :(
        if syntaxcheck and (i <> line) then begin
          F[0].Start := 0;
          F[0].Length := Length(s);
          F[0].Font := unitAgentForm.err;
          errorinfo[i] := e.message;
        end;
        if i <> line then Exit;
      end;
    end;

    if syntaxcheck then begin
      err := SyntaxCheckTokenSet(t);
      j := 0;
      while err[j].errortype <> cetNone do begin
        if err[j].errortype = cetTooManyParams then errorinfo[i] := errorinfo[i] + 'too many parameters'
        else if err[j].errortype = cetNotEnoughParams then errorinfo[i] := errorinfo[i] + 'not enough parameters.'
        else if err[j].errortype = cetFirstTokenNotCommand then errorinfo[i] := errorinfo[i] + 'must start with command'
        else if err[j].errortype = cetTypeMismatch then errorinfo[i] := errorinfo[i] + 'mismatch on token ' + IntToStr(err[j].tokenno) + ' expected ' + typetostring(err[j].expected) + ' got ' + typetostring(err[j].actual);
        Inc(j);
      end;
      if (errorinfo[i] <> '') and (i <> line) then begin
        F[0].Start := 0;
        F[0].Length := Length(s);
        F[0].Font := unitAgentForm.err;
        Exit;
      end;
    end;

    // we now cycle through all the tokens in the token set
    // if we encounter whitespace characters, we process them
    // FIXME: does Creatures like multiple whitespaces, hmm? we might have to flag that on syntax :)
    //        (later note: looking at the c2e source, it strips extra whitespaces)

    // SPEED is of the essence here ... the tokeniser is slow enough
    j := 1; k := 0; p := 1;
    while j <= Length(s) do begin
      if (s[j] = ' ') or (s[j] = #9) then begin
        // not a token
        F[k].Start := j - 1;
        F[k].Length := 1;
        F[k].Font := black; // doesn't really matter
        Inc(k);
      end else begin
        if p > t.notokens then begin ColorCode := False; ErrorDialog(13, True); end; // (note that if the tokeniser messes up then this all trashes)
        F[k].Start := j - 1; // make sure it aligns
        if t.tokens[p].stringvalue[1] = '*' then begin
          F[k].Length := Length(s) - j;
          F[k].Font := green;
          Exit;
        end;

        F[k].Length := Length(t.tokens[p].stringvalue);
        if t.tokens[p].iskeyword then F[k].Font := blue
        else if t.tokens[p].tokentype = ctString then F[k].Font := purple
        else if t.tokens[p].tokentype = ctByteString then F[k].Font := purple
        else F[k].Font := black;
        j := j + Length(t.tokens[p].stringvalue) - 1;
        Inc(p); Inc(k);
      end;
      Inc(j);
    end;
  except
    on e: Exception do errorinfo[i] := 'internal error (' + e.message + ') - report to Fuzzie'; // meh x 5000
  end; finally
    t.Free;
  end;
end;

procedure TAgentForm.SyntaxMemoChange(Sender: TObject);
begin
  modified := true;
end;

procedure TAgentForm.SyntaxMemoKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  i, line, n: Integer;
  t: TTokenInfo;
  s: string;
begin
  SyntaxMemoEnter(Sender);
  if (Key = VK_UP) or (Key = VK_DOWN) then LineChangeSync;
  if key = VK_RETURN then begin
    // fixme: the auto indent code is as buggy as hell
    if not autoindent then Exit;
    // automatic indenting
    // fixme: this should be exited if we're on overwrite mode (not insert)

    line := SendMessage(SyntaxMemo.Handle, EM_LINEFROMCHAR, SyntaxMemo.SelStart, 0);
    internal_cmd := ctCommand;

    n := 0;

    for i := 0 to (line - 1) do begin
      try
        t.SlurpToken(SyntaxMemo.Lines[i]);
      except
        on Exception do Continue;
      end;

      if t.iskeyword then begin
        s := LowerCase(t.stringvalue);

        if s = 'iscr' then n := n + 1;
        if s = 'rscr' then n := n + 1;
        if s = 'scrp' then n := n + 1;
        if s = 'endm' then n := n - 1;

        if s = 'doif' then n := n + 1;
        if s = 'endi' then n := n - 1;

        if s = 'subr' then n := n + 1;
        if s = 'enum' then n := n + 1;
        if s = 'next' then n := n - 1;

        if s = 'loop' then n := n + 1;
        if s = 'untl' then n := n - 1;
      end;
    end;

    s := '';
    for i := 1 to n do begin
      s := s + '  ';
    end;
    SyntaxMemo.Lines[line] := s + SyntaxMemo.Lines[line]; // kludge?

    internal_cmd := ctAnything; // reset
  end;
end;

procedure TAgentForm.LineChangeSync;
var
  line: Integer;
  t: TTokenInfo;
begin
  SyntaxMemo.Refresh;

  line := SendMessage(SyntaxMemo.Handle, EM_LINEFROMCHAR, SyntaxMemo.SelStart, 0);

  t := TTokenInfo.Create;
  try
    internal_cmd := ctCommand;
    t.SlurpToken(LowerCase(Trim(SyntaxMemo.Lines[line])));

    if t.iskeyword then begin
      CAOSRefForm.ComboBox1.ItemIndex := 0;
      CAOSRefForm.ComboBox1Change(Self);
      CAOSRefForm.ComboBox2.ItemIndex := CAOSRefForm.ComboBox2.Items.IndexOf(LowerCase(t.stringvalue));
      CAOSRefForm.ComboBox2Change(Self);
    end;

    internal_cmd := ctAnything; // reset it
  except
    on Exception do begin t.Free; Exit; end;
  end;

  t.Free;
end;

procedure TAgentForm.Indent1Click(Sender: TObject);
begin
  ErrorDialog(15, True);
end;

procedure TAgentForm.InjectintoCreatures1Click(Sender: TObject);
var
  install, remove, events: TStringList;
begin
  // FIXME: this injection method only works with Creatures 3/Adventures
  CommLink.Connect;
  if not CommLink.Connected then begin ErrorDialog(9, True); Exit; end;

  install := TStringList.Create; remove := TStringList.Create; events := TStringList.Create;
  SplitScriptStrings(SyntaxMemo.Lines, install, remove, events);
  CommLink.Transaction := install.GetText;
  CommLink.Transaction := events.GetText;
  install.Free; remove.Free; events.Free;

  CommLink.Disconnect;
end;

procedure TAgentForm.View1Click(Sender: TObject);
begin
  ColorCoding1.Checked := ColorCode;
  CAOSRefSync1.Checked := CAOSRefSync;
  SyntaxChecking1.Checked := SyntaxCheck;
  ShowSyntaxArea1.Checked := SyntaxAreaVisible;
  AutoIndent1.Checked := autoindent;
end;

procedure TAgentForm.ColorCoding1Click(Sender: TObject);
begin
  ColorCode := not ColorCode;
  SyntaxMemo.Refresh;
end;

procedure TAgentForm.CAOSRefSync1Click(Sender: TObject);
begin
  CAOSRefSync := not CAOSRefSync;
end;

procedure TAgentForm.SyntaxChecking1Click(Sender: TObject);
begin
  SyntaxCheck := not SyntaxCheck;
  SyntaxMemo.Refresh;
end;

procedure TAgentForm.Undo1Click(Sender: TObject);
begin
  SendMessage(SyntaxMemo.Handle, WM_UNDO, 0, 0);
end;

procedure TAgentForm.Delete1Click(Sender: TObject);
begin
  SyntaxMemo.SelText := '';
end;

procedure TAgentForm.SelectAll1Click(Sender: TObject);
begin
  SyntaxMemo.SelectAll;
end;

procedure SplitScriptStrings(src, install, remove, events: TStrings);
var
  t: TTokenInfo;
  i, location: Integer; // 0 = install, 1 = remove, 2 = events
  s: string;
begin
  location := 0; // we start on install usually

  install.Clear; remove.Clear; events.Clear;

  t := TTokenInfo.Create;
  for i := 0 to (src.Count - 1) do begin
    s := Trim(src[i]);
    if s = '' then Continue;

    try
      t.SlurpToken(LowerCase(s));
      if t.stringvalue = 'iscr' then begin location := 0; Continue; end
      else if t.stringvalue = 'rscr' then begin location := 1; Continue; end
      else if t.stringvalue = 'scrp' then location := 2
      else if t.stringvalue = 'endm' then begin location := -1; Continue; end;
    except
      on Exception do;
    end;

    if location = -1 then begin ErrorDialog(16, True); end;

    if location = 0 then install.Add(s)
    else if location = 1 then remove.Add(s)
    else if location = 2 then events.Add(s);
  end;
  t.Free;
end;

procedure TAgentForm.CompiletoPRAY1Click(Sender: TObject);
var
  b: TPRAYFile;
  stream: TStream;
  noscrips, F, i: Integer;
  s, q, q2, sounddir, imagedir: string;
  install, remove, events, praysource: TStringList;
  r: TRegistry;

  t: TTokenInfo;
begin
  r := TRegistry.Create;
  r.RootKey := HKEY_LOCAL_MACHINE;
  if (CommLink.GameName = 'Creatures 2') or (not r.OpenKey('Software\CyberLife Technology\' + CommLink.GameName, False)) then begin
    Application.MessageBox(PChar('The game ' + CommLink.GameName + ' doesn''t support PRAY!'), 'Zeus', MB_OK + MB_ICONWARNING);
    Exit;
  end;
  imagedir := r.ReadString('Images Directory');
  sounddir := r.ReadString('Sounds Directory');
  r.Free;

  praysource := TStringList.Create;
  with TPRAYCompileForm.Create(Self) do begin
    t := TTokenInfo.Create;
    internal_cmd := ctCommand;
    for i := 0 to (SyntaxMemo.Lines.Count - 1) do begin
      s := Trim(SyntaxMemo.Lines[i]);
      try s := t.SlurpToken(s); except on Exception do Continue; end;
      q := LowerCase(t.stringvalue);
      if (((q = 'snde') or (q = 'sndl')) or (q = 'sndc')) or (q = 'sndq') then begin
        try s := t.SlurpToken(s); except on Exception do Continue; end;
        q := Copy(t.stringvalue, 2, Length(t.stringvalue) - 2); // remove the quotes
        q := sounddir + q + '.wav';
        if FileExists(q) then
          if ListBox1.Items.IndexOf(q) = -1 then
            ListBox1.Items.Add(q);
      end else if ((q = 'new: simp') or (q = 'new: vhcl')) or (q = 'new: comp') then begin
        try
          s := t.SlurpToken(s); s := t.SlurpToken(s);
          s := t.SlurpToken(s); s := t.SlurpToken(s);
        except
          on Exception do Continue;
        end;
        q := Copy(t.stringvalue, 2, Length(t.stringvalue) - 2); // remove the quotes
        q2 := imagedir + q + '.c16';
        if not FileExists(q2) then q2 := imagedir + q + '.s16';
        if FileExists(q2) then
          if ListBox1.Items.IndexOf(q2) = -1 then
            ListBox1.Items.Add(q2);
      end else if q = 'gall' then begin
        try s := t.SlurpToken(s); except on Exception do Continue; end;
        q := Copy(t.stringvalue, 2, Length(t.stringvalue) - 2); // remove the quotes
        q2 := imagedir + q + '.c16';
        if not FileExists(q2) then q2 := imagedir + q + '.s16';
        if FileExists(q2) then
          if ListBox1.Items.IndexOf(q2) = -1 then
            ListBox1.Items.Add(q2);
      end;
      q := ''; s := '';
    end;
    t.Free;
    internal_cmd := ctAnything; // reset
    if ShowModal = IDCANCEL then begin Free; Exit; end;
    if SaveDialog1.Execute then begin
      b := TPRAYFile.Create;
      b.noblocks := 0;
      b.blocks[0] := TStandardPRAYBlock.Create;
      with b.blocks[0] as TStandardPRAYBlock do begin
        header.PRAYtype := 'AGNT';
        Move(PChar(Edit1.Text)^, header.name, Length(Edit1.Text) + 1);
        if CheckBox1.Checked then header.flags := 0
        else header.flags := 1;
        noints := 0;
        nostrs := 0;

        if CheckBox3.Checked then begin
          Inc(noints);
          intnames[noints] := 'Camera X';
          intvalues[noints] := SpinEdit1.Value;
          Inc(noints);
          intnames[noints] := 'Camera Y';
          intvalues[noints] := SpinEdit2.Value;
        end;

        Inc(noints);
        intnames[noints] := 'Agent Bioenergy Value';
        intvalues[noints] := SpinEdit3.Value;

        Inc(noints);
        intnames[noints] := 'Agent Type';
        intvalues[noints] := 0;

        Inc(noints);
        intnames[noints] := 'Dependency Count';
        intvalues[noints] := ListBox1.Items.Count;

        for i := 1 to ListBox1.Items.Count do begin
          s := ListBox1.Items[i - 1];
          Inc(nostrs);
          strnames[nostrs] := 'Dependency ' + IntToStr(i);
          strvalues[nostrs] := ExtractFileName(s);
          Inc(noints);
          intnames[noints] := 'Dependency Category ' + IntToStr(i);
          s := LowerCase(ExtractFileExt(s));
          if s = '.catalogue' then intvalues[noints] := 7
          else if s = '.c16' then intvalues[noints] := 2
          else if s = '.s16' then intvalues[noints] := 2
          else if s = '.blk' then intvalues[noints] := 6
          else if s = '.gen' then intvalues[noints] := 3
          else if s = '.gno' then intvalues[noints] := 3
          else if s = '.wav' then intvalues[noints] := 1
          else intvalues[noints] := 0;
        end;

        Inc(nostrs);
        strnames[nostrs] := 'Agent Animation File';
        strvalues[nostrs] := ExtractFileName(Edit2.Text);
        Inc(nostrs);
        strnames[nostrs] := 'Agent Animation Gallery';
        strvalues[nostrs] := ExtractNoExt(Edit2.Text);
        Inc(nostrs);
        strnames[nostrs] := 'Agent Animation String';
        strvalues[nostrs] := Edit3.Text;

        install := TStringList.Create;
        remove := TStringList.Create;
        events := TStringList.Create;

        SplitScriptStrings(SyntaxMemo.Lines, install, remove, events);

        // process strings
        Inc(nostrs);
        strnames[nostrs] := 'Script 1';
        strvalues[nostrs] := string(install.GetText);

        noscrips := 1;

        install.Clear; // we'll now use install for holding our data
        t := TTokenInfo.Create;
        for i := 0 to (events.Count - 1) do begin
          s := events[i]; // already trimmed etc by SplitScriptStrings

          t.SlurpToken(LowerCase(s));
          if t.stringvalue = 'scrp' then begin
            if noscrips > 1 then begin
              Inc(nostrs);
              strnames[nostrs] := 'Script ' + IntToStr(noscrips);
              strvalues[nostrs] := string(install.GetText);
            end;
            Inc(noscrips);
          end;

          install.Add(s);
        end;
        if noscrips > 1 then begin
          Inc(nostrs);
          strnames[nostrs] := 'Script ' + IntToStr(noscrips);
          strvalues[nostrs] := string(install.GetText);
        end;
        t.Free;

        Inc(nostrs);
        strnames[nostrs] := 'Remove Script';
        strvalues[nostrs] := string(remove.GetText);

        Inc(noints);
        intnames[noints] := 'Script Count';
        intvalues[noints] := noscrips;

        install.Free;
        remove.Free;
        events.Free;

        praysource.Add('# PRAY Source File (generated by Zeus)');
        praysource.Add('# WARNING: this PRAY file may not compile properly using Creature Labs tools...');
        praysource.Add('');
        praysource.Add('"en-gb" # Targetted for british :) (US english etc)');
        praysource.Add('');
        praysource.Add('group AGNT "' + Edit1.Text + '"');

        for i := 1 to noints do begin
          praysource.Add('"' + intnames[i] + '" ' + IntToStr(intvalues[i]));
        end;
        for i := 1 to nostrs do begin
          praysource.Add('"' + strnames[i] + '" "' + strvalues[i] + '"');
        end;
      end;

      for i := 1 to ListBox1.Items.Count do begin
        s := ListBox1.Items[i - 1];
        b.noblocks := b.noblocks + 1;
        b.blocks[b.noblocks] := TNullPRAYBlock.Create;
        with b.blocks[b.noblocks] as TNullPRAYBlock do begin
          ZeroMemory(@header.name, 128);
          F := FileOpen(s, fmOpenRead);
          s := ExtractFileName(s);
          Move(PChar(s)^, header.name, Length(s));
          header.size := FileSeek(F, 0, 2);
          header.usize := header.size;
          header.praytype := 'FILE';
          FileSeek(F, 0, 0); // back to the beginning
          GetMem(data, header.size);
          FileRead(F, data^, header.size);
          header.flags := 0;
        end;
        FileClose(F);
      end;

      stream := TFileStream.Create(SaveDialog1.Filename, fmCreate);
      b.SaveToFile(stream);
      stream.Free;
      b.Free;
      if CheckBox4.Checked then praysource.SaveToFile(SaveDialog1.Filename + '.ps');
      Free;
    end; // SaveDialog1.Execute
  end;
  praysource.Free;
end;

procedure TAgentForm.InjectRemovalScript1Click(Sender: TObject);
var
  install, remove, events: TStringList;
begin
  // FIXME: this injection method only works with Creatures 3/Adventures
  CommLink.Connect;
  if not CommLink.Connected then begin ErrorDialog(9, True); Exit; end;

  install := TStringList.Create; remove := TStringList.Create; events := TStringList.Create;
  SplitScriptStrings(SyntaxMemo.Lines, install, remove, events);
  CommLink.Transaction := remove.GetText;
  install.Free; remove.Free; events.Free;

  CommLink.Disconnect;
end;

procedure TAgentForm.Find1Click(Sender: TObject);
begin
  FindDialog1.Execute;
end;

procedure TAgentForm.Replace1Click(Sender: TObject);
begin
  ReplaceDialog1.Execute;
end;

procedure TAgentForm.FindDialog1Find(Sender: TObject);
var
  Buff, P, FT: PChar;
  BuffLen: Integer;
begin
  with Sender as TFindDialog do begin
    GetMem(FT, Length(FindText) + 1);
    StrPCopy(FT, FindText);
    BuffLen := SyntaxMemo.GetTextLen + 1;
    GetMem(Buff, BuffLen);
    SyntaxMemo.GetTextBuf(Buff, BuffLen);
    P := Buff + SyntaxMemo.SelStart + SyntaxMemo.SelLength;
    P := StrPos(P, FT);
    if P = nil then MessageBeep($FFFFFFFF)
    else begin
      SyntaxMemo.SelStart := P - Buff;
      SyntaxMemo.SelLength := Length(FindText);
    end;
    FreeMem(FT);
    FreeMem(Buff);
  end;
end;

procedure TAgentForm.ReplaceDialog1Replace(Sender: TObject);
begin
  with Sender as TReplaceDialog do begin
    while True do begin
      if SyntaxMemo.SelText <> FindText then
        FindDialog1Find(Sender);
      if SyntaxMemo.SelLength = 0 then Break;
      SyntaxMemo.SelText := ReplaceText;
      if not (frReplaceAll in Options) then Break;
    end;
  end;
end;

// NOTE: this function crashes if you execute it from the Delphi 3 environment
// I'm not sure why ... it seems to work fine on its own though.

// Later Note: it works fine if you set a breakpoint in the loop and step
//             through every step ... grah
procedure TAgentForm.SyntaxCheck1Click(Sender: TObject);
var
  t: TTokenSet;
  i, j, p: Integer;
  x, s: string;
  err: TCAOSErrorData;
begin
  ListBox1.Items.Clear;
  SyntaxAreaVisible := True;
  ListBox1.Visible := SyntaxAreaVisible;

  for i := 0 to (SyntaxMemo.Lines.Count - 1) do begin
    if i = 9 then
      Application.ProcessMessages;

    s := SyntaxMemo.Lines[i];
    s := LowerCase(Trim(s));

    // quick checks for speed
    if Length(s) = 0 then Continue;
    if s[1] = '*' then Continue;

    t := TTokenSet.Create;
    try
      try
        t.SlurpTokenSet(s);
      except
        on e: Exception do begin
          ListBox1.Items.AddObject('line ' + IntToStr(i) + ': ' + e.message, TObject(i));
          Continue;
        end;
      end;

      err := SyntaxCheckTokenSet(t);
      j := 0;
      x := '';
      while err[j].errortype <> cetNone do begin
        if err[j].errortype = cetTooManyParams then x := x + 'too many parameters'
        else if err[j].errortype = cetNotEnoughParams then x := x + 'not enough parameters.'
        else if err[j].errortype = cetFirstTokenNotCommand then x := x + 'must start with command'
        else if err[j].errortype = cetTypeMismatch then x := x + 'mismatch on token ' + IntToStr(err[j].tokenno) + ' expected ' + typetostring(err[j].expected) + ' got ' + typetostring(err[j].actual);
        Inc(j);
        if j > 1000 then begin ErrorDialog(17, True); Exit; end;
      end;

      if x <> '' then begin
        ListBox1.Items.AddObject('line ' + IntToStr(i) + ': ' + x, TObject(i));
        for j := 1 to t.notokens do
          if t.tokens[j].iskeyword then
            with t.tokens[j].keywordinfo do begin
              s := '   ''' + name + ''' (' + typetostring(ktype) + '): ';
              if noparams = 0 then s := s + '(no parameters)'
              else for p := 1 to noparams do begin
                if p > 1 then s := s + ', ';
                s := s + paramnames[p] + ' (' + typetostring(paramtypes[p]) + ')';
              end;
              ListBox1.Items.Add(s);
            end;
      end;
    finally
      t.Free;
    end;
  end;
end;

procedure TAgentForm.FormActivate(Sender: TObject);
begin
  ListBox1.Visible := SyntaxAreaVisible;
end;

procedure TAgentForm.ShowSyntaxArea1Click(Sender: TObject);
begin
  SyntaxAreaVisible := not SyntaxAreaVisible;
  ListBox1.Visible := SyntaxAreaVisible;
end;

procedure TAgentForm.ListBox1Click(Sender: TObject);
var
  line: Integer;
begin
  if ListBox1.ItemIndex = -1 then Exit;
  line := Integer(ListBox1.Items.Objects[ListBox1.ItemIndex]);
  if line = 0 then Exit;

  SyntaxMemo.SelStart := SendMessage(SyntaxMemo.Handle, EM_LINEINDEX, line, 0);
  SyntaxMemo.SelLength := Length(SyntaxMemo.Lines[line]);
  SyntaxMemo.SetFocus;
  SyntaxMemo.SelStart := SendMessage(SyntaxMemo.Handle, EM_LINEINDEX, line, 0);
  SyntaxMemo.Refresh; // paint the color-coding
end;

procedure TAgentForm.SyntaxMemoEnter(Sender: TObject);
var
  line: Integer;
begin
  line := SendMessage(SyntaxMemo.Handle, EM_LINEFROMCHAR, SyntaxMemo.SelStart, 0);
  if errorinfo[line] <> '' then ZeusMainForm.StatusBar1.SimpleText := 'Line ' + IntToStr(line) + ', ' + errorinfo[line] + '.'
  else ZeusMainForm.StatusBar1.SimpleText := 'Line ' + IntToStr(line) + ', no problems.';
end;

procedure TAgentForm.SyntaxMemoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  SyntaxMemoEnter(Sender);
  LineChangeSync;
end;

procedure TAgentForm.AutoIndent1Click(Sender: TObject);
begin
  autoindent := not autoindent;
end;

procedure TAgentForm.InjectEventScripts1Click(Sender: TObject);
var
  install, remove, events: TStringList;
begin
  // FIXME: this injection method only works with Creatures 3/Adventures
  CommLink.Connect;
  if not CommLink.Connected then begin ErrorDialog(9, True); Exit; end;

  install := TStringList.Create; remove := TStringList.Create; events := TStringList.Create;
  SplitScriptStrings(SyntaxMemo.Lines, install, remove, events);
  CommLink.Transaction := events.GetText;
  install.Free; remove.Free; events.Free;

  CommLink.Disconnect;
end;

initialization
  RegisterEditor('Agent Files', 'cos', 'Agent Editor', TAgentForm);

  SyntaxAreaVisible := False;

  // setup fonts ...
  globalfontname := 'Courier New'; globalfontsize := 8;

  blue := TFont.Create; blue.Color := clBlue; blue.Name := globalfontname; blue.Size := globalfontsize;
  green := TFont.Create; green.Color := clGreen; green.Name := globalfontname; green.Size := globalfontsize;
  red := TFont.Create; red.Color := clRed; red.Name := globalfontname; red.Size := globalfontsize;
  err := TFont.Create; err.Color := clRed; err.Name := globalfontname; err.Style := [fsBold]; err.Size := globalfontsize;
  purple := TFont.Create; purple.Color := clPurple; purple.Name := globalfontname; purple.Size := globalfontsize;
  black := TFont.Create; black.Color := clBlack; black.Name := globalfontname; black.Size := globalfontsize;
  bold := TFont.Create; bold.Style := [fsBold]; bold.Name := globalfontname; bold.Size := globalfontsize;
finalization
  blue.Free; green.Free; red.Free; err.Free;
  purple.Free; black.Free; bold.Free;
end.
