unit unitBLKCreator;

// TBLKCreatorForm : BLK Creator

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Buttons, Spin;

type
  TBLKCreatorForm = class(TForm)
    Label1: TLabel; Label2: TLabel; Label3: TLabel; Label4: TLabel; SpinEdit1: TSpinEdit; SpinEdit2: TSpinEdit;
    RadioGroup1: TRadioGroup; RadioGroup2: TRadioGroup; Edit1: TEdit; SpeedButton1: TSpeedButton;
    Button1: TButton; Button2: TButton; OpenDialog1: TOpenDialog; SaveDialog1: TSaveDialog;

    procedure RadioGroup1Click(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Edit1Change(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    procedure DoBLKCreate;
  end;

var
  BLKCreatorForm: TBLKCreatorForm;

implementation

uses BLKClasses, unitMapEditor;

{$R *.DFM}

procedure TBLKCreatorForm.RadioGroup1Click(Sender: TObject);
begin
  RadioGroup2.Enabled := (RadioGroup1.ItemIndex = 0) or (RadioGroup1.ItemIndex = 4);

  if RadioGroup1.ItemIndex = 2 then ClientHeight := 195
  else ClientHeight := 145;
end;

procedure TBLKCreatorForm.SpeedButton1Click(Sender: TObject);
begin
  if RadioGroup1.ItemIndex = 0 then begin
    OpenDialog1.DefaultExt := 'bmp';
    OpenDialog1.Filter := 'Bitmap Files (*.bmp)|*.bmp';
  end else if RadioGroup1.ItemIndex = 4 then begin
    OpenDialog1.DefaultExt := 'spr';
    OpenDialog1.Filter := 'C1 Sprite Files (*.spr)|*.spr';
  end else begin
    OpenDialog1.DefaultExt := 's16';
    OpenDialog1.Filter := 'Bitmap Files (*.s16)|*.s16';
  end;

  if OpenDialog1.Execute then
    Edit1.Text := OpenDialog1.Filename;
end;

procedure TBLKCreatorForm.FormCreate(Sender: TObject);
begin
  Edit1.Text := '';
  RadioGroup1.ItemIndex := 0;
  RadioGroup2.ItemIndex := 0;
  Button1.Enabled := False;
end;

procedure TBLKCreatorForm.Edit1Change(Sender: TObject);
begin
  Button1.Enabled := FileExists(Edit1.Text);
end;

procedure TBLKCreatorForm.DoBLKCreate;
var
  s1, s2: TS16File;
  b: TBitmap;
begin
  if RadioGroup1.ItemIndex > 2 then begin
    Application.MessageBox('Sorry, Zeus doesn''t support converting this source image type to BLK right now. However, support may be available in the next release.', 'Zeus', MB_OK + MB_ICONERROR);
    Exit;
  end;

  SaveDialog1.InitialDir := backdir;
  if not SaveDialog1.Execute then Exit;

  // oke ... 0 = BMP, 1 = S16, 2 = S16 blocks
  // 2 is the easiest =) we just load it in, ask the user to tell us the width/height, then save it as a BLK
  // 1 is quite easy ... we just grab the data and dump it into 128x128 blocks, then save it as a BLK
  // 0 is hard ... we need to convert the data (into 555/565 if needed also) then do number 1

  if RadioGroup1.ItemIndex = 2 then begin
    s1 := TS16File.Create;
    s1.LoadFromS16(Edit1.Text);
    WriteS16AsBLK(SaveDialog1.Filename, s1, SpinEdit1.Value, SpinEdit2.Value);
    s1.FreeData; s1.Free;
  end;

  if RadioGroup1.ItemIndex = 1 then begin
    s1 := TS16File.Create;
    s1.LoadFromS16(Edit1.Text);
    s2 := TranslateData(s1.frames[1].data, s1.frames[1].width div 128, s1.frames[1].height div 128);
    s1.FreeData;
    if RadioGroup2.ItemIndex = 1 then s2.ConvertTo565;
    WriteS16AsBLK(SaveDialog1.Filename, s2, s1.frames[1].width div 128, s1.frames[1].height div 128);
    s1.Free;
    s2.FreeData;
    s2.Free;
  end;

  if RadioGroup1.ItemIndex = 0 then begin
    b := TBitmap.Create;
    b.LoadFromFile(Edit1.Text);
    if ((b.Width mod 128) <> 0) or ((b.Height mod 128) <> 0) then begin
      b.Free;
      Application.MessageBox('The dimensions (width+height) of the bitmap used for creating a BLK file must be multiples of 128! Try using an "Enlarge Canvas" option in your graphics program.', 'Zeus', MB_OK + MB_ICONERROR);
      Exit;
    end;
    s1 := TS16File.Create;
    s1.noframes := 1;
    s1.frames[1] := TS16Frame.Create;
    s1.is_565 := false;
    s1.frames[1].is_565 := false;

    s1.frames[1].LoadFromBMP(b, Canvas);
    if RadioGroup2.ItemIndex = 1 then s1.ConvertTo565;
    b.Free;
    s2 := TranslateData(s1.frames[1].data, s1.frames[1].width div 128, s1.frames[1].height div 128);
    s2.is_565 := s1.is_565;
    // if RadioGroup2.ItemIndex = 1 then s2.ConvertTo565;
    s1.FreeData;
    WriteS16AsBLK(SaveDialog1.Filename, s2, s1.frames[1].width div 128, s1.frames[1].height div 128);
    s1.Free;
    s2.Free;
  end;
end;

procedure TBLKCreatorForm.Button1Click(Sender: TObject);
begin
  DoBLKCreate;
end;

end.
