unit unitPRAYDecompiler;

// TPRAYDecompilerForm: PRAY/COB Viewer/Editor

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, AgentClasses, unitChildForm, ComCtrls, StdCtrls, ExtCtrls, unitErrorHandler,
  ImgList;

type
  TPRAYDecompilerForm = class(TMDIChildForm)
    Memo1: TMemo; ListView1: TListView; ImageList1: TImageList; Panel1: TPanel;
    Button1: TButton; Button2: TButton; SaveDialog1: TSaveDialog;

    procedure ListView1Change(Sender: TObject; Item: TListItem; Change: TItemChange);
    procedure Button1Click(Sender: TObject);
    procedure ListView1Edited(Sender: TObject; Item: TListItem; var S: String);
    procedure ListView1Editing(Sender: TObject; Item: TListItem; var AllowEdit: Boolean);
  public
    prayfile: Boolean; pray: TPRAYFile; cob: TCOBFile;

    procedure ReSync;

    function IsFileBased: Boolean; override;
    procedure New; override;
    procedure Open(fname: string); override;
    procedure Save; override;
  end;

var
  PRAYDecompilerForm: TPRAYDecompilerForm;

implementation

uses Unit1;

{$R *.DFM}

function TPRAYDecompilerForm.IsFileBased: Boolean;
begin
  Result := True;
end;

procedure TPRAYDecompilerForm.New;
begin
  inherited;

  ErrorDialog(13, True);
end;

procedure TPRAYDecompilerForm.Open(fname: string); // open ze existing file
var
  header: array[0..3] of Char;
  stream: TStream;
  F: Integer;
begin
  inherited;
  stream := TFileStream.Create(filename, fmOpenRead);
  stream.Read(header, 4);
  if header = 'PRAY' then prayfile := true
  else prayfile := false;
  stream.Seek(0, 0);

  fname := filename;

  if prayfile then begin
    pray := TPRAYFile.Create;
    pray.LoadFromFile(stream);

    for F := 0 to pray.noblocks do begin
      with ListView1.Items.Add do
        Caption := pray.blocks[F].header.name;
    end;
  end else begin
    cob := TCOBFile.Create;
    cob.ReadCOBFile(stream);

    for F := 1 to cob.noblocks do begin
      with ListView1.Items.Add do
        Caption := cob.cobblocks[F].Caption;
    end;
  end;

  Caption := 'ZeusPRAY - [' + ExtractFileName(filename) + ']';

  stream.Free;
end;

procedure TPRAYDecompilerForm.Save;
var
  stream: TFileStream;
begin
  stream := TFileStream.Create(filename, fmCreate);
  if prayfile then pray.SaveToFile(stream)
  else cob.WriteCobFile(stream);
  stream.Free;
end;

procedure TPRAYDecompilerForm.ReSync;
begin
  Memo1.Lines.Clear;
  ListView1.Items.Clear;
end;

procedure TPRAYDecompilerForm.ListView1Change(Sender: TObject;
  Item: TListItem; Change: TItemChange);
var
  str: TStringList;
  k, j: Integer;
begin
  Button1.Enabled := False;

  str := TStringList.Create;

  if ListView1.Selected = nil then Exit;

  if prayfile then begin
    str.Add('(-');
    str.Add(' CEngine PRAY File');
    str.Add(' Generated from ' + ExtractFileName(filename) + ' by ZeusPRAY');
    str.Add(' http://www.warpedgames.com/');
    str.Add('-)');
    str.Add('');
    str.Add('"en-gb" # Targetted for british :) (US english etc)');
    str.Add('');

    k := ListView1.Selected.Index;

    str.Add('group ' + pray.blocks[k].header.PRAYType + ' "' + pray.blocks[k].header.name + '"');
    if (pray.blocks[k].header.flags and 1) = 1 then str.Add('# this block was compressed');

    if pray.blocks[k] is TStandardPRAYBlock then with pray.blocks[k] as TStandardPRAYBlock do begin
      for j := 1 to noints do str.Add('"' + intnames[j] + '" ' + IntToStr(intvalues[j]));
      for j := 1 to nostrs do str.Add('"' + strnames[j] + '" "' + strvalues[j] + '"');
    end else if pray.blocks[k].header.PRAYType = 'FILE' then begin
      Button1.Enabled := True;
      str.Add('# file block - use Extract button to extract');
    end else begin
      str.Add('# unknown block type, data may be missing');
    end;
  end else begin
    str.Add('(-');
    if cob.c1format then
      str.Add(' Creatures 1 COB File')
    else
      str.Add(' Creatures 2 COB File');
    str.Add(' Generated from ' + ExtractFileName(filename) + ' by ZeusPRAY');
    str.Add(' http://www.warpedgames.com/');
    str.Add('-)');
    str.Add('');
    str.Add('"en-gb" # Targetted for british :) (US english etc)');
    str.Add('');

    if cob.cobblocks[ListView1.Selected.Index + 1] is TFileBlock then with cob.cobblocks[ListView1.Selected.Index + 1] as TFileBlock do begin
      str.Add('group AUTH "' + FileName + '"');
      str.Add('"FileType" ' + IntToStr(FileType));
      str.Add('"FileSize" ' + IntToStr(FileSize));
    end else if cob.cobblocks[ListView1.Selected.Index + 1] is TAgentBlock then with cob.cobblocks[ListView1.Selected.Index + 1] as TAgentBlock do begin
      str.Add('group AGNT "' + AgentName + '"');
      str.Add('"Expires Day" ' + IntToStr(Integer(header.UseByDay)));
      str.Add('"Expires Month" ' + IntToStr(Integer(header.UseByMonth)));
      str.Add('"Expires Year" ' + IntToStr(header.UseByYear));
      str.Add('"Quantity Remaining" ' + IntToStr(Integer(header.QuantityRemaining)));
      str.Add('"Last Usage" ' + IntToStr(Integer(header.LastUsage)));
      str.Add('"Reuse Interval" ' + IntToStr(header.ReuseInterval));
      str.Add('"Thumbnail Width" ' + IntToStr(Integer(ThumbnailWidth)));
      str.Add('"Thumbnail Height" ' + IntToStr(Integer(ThumbnailHeight)));
      str.Add('"Agent Description" "' + AgentDescription + '"');
      str.Add('"Install Script" "' + InstallScript + '"');
      str.Add('"Removal Script" "' + RemovalScript + '"');
      str.Add('"Number of Events" ' + IntToStr(NumberOfEvents));
      for k := 1 to NumberOfEvents do begin
        str.Add('"Event Script ' + IntToStr(k) + '" "' + EventScripts[k] + '"');
      end;
      str.Add('"Number of File Dependancies" ' + IntToStr(NumberOfDeps));
      for k := 1 to NumberOfDeps do begin
        str.Add('"File Dependancy ' + IntToStr(k) + ' Name" "' + FileDeps_name[k] + '"');
        str.Add('"File Dependancy ' + IntToStr(k) + ' Type" ' + IntToStr(FileDeps_type[k]));
      end;
    end else if cob.cobblocks[ListView1.Selected.Index + 1] is TAuthBlock then with cob.cobblocks[ListView1.Selected.Index + 1] as TAuthBlock do begin
      str.Add('group AUTH "Author Information"');
      str.Add('"Day" ' + IntToStr(Integer(header.Day)));
      str.Add('"Month" ' + IntToStr(Integer(header.Month)));
      str.Add('"Year" ' + IntToStr(header.Year));
      str.Add('"Version" ' + IntToStr(Integer(header.Version)));
      str.Add('"Revision" ' + IntToStr(Integer(header.Revision)));
      str.Add('"Author Name" "' + AuthorName + '"');
      str.Add('"Author E-Mail" "' + AuthorEMail + '"');
      str.Add('"Author URL" "' + AuthorURL + '"');
      str.Add('"Author Comments" "' + AuthorComments + '"');
    end else begin
      str.Add('# unknown block type, data may be missing');
    end;
  end;

  Memo1.Lines.Assign(str);
  str.Free;
end;

procedure TPRAYDecompilerForm.Button1Click(Sender: TObject);
var
  F: Integer;
begin
  inherited;

  SaveDialog1.Filename := pray.blocks[ListView1.Selected.Index].header.name;
  if SaveDialog1.Execute then begin
    F := FileCreate(SaveDialog1.Filename);
    with pray.blocks[ListView1.Selected.Index] as TNullPRAYBlock do begin
      FileWrite(F, data^, header.usize);
    end;
    FileClose(F);
  end;
end;

procedure TPRAYDecompilerForm.ListView1Edited(Sender: TObject; Item: TListItem; var S: String);
begin
  if Length(s) > 127 then begin
    Application.MessageBox('The length of a block name is restricted to 127 characters.', 'Zeus', MB_OK + MB_ICONERROR);
    item.caption := pray.blocks[item.Index].header.name;
    Exit;
  end;
  StrPCopy(pray.blocks[item.Index].header.name, s);
  ListView1Change(Self, nil, TItemChange(0));
end;

procedure TPRAYDecompilerForm.ListView1Editing(Sender: TObject; Item: TListItem; var AllowEdit: Boolean);
begin
  if prayfile then AllowEdit := True
  else begin Application.MessageBox('Sorry, you cannot edit block names of COB files.', 'Zeus', MB_OK + MB_ICONERROR); AllowEdit := False; end; // fixme: can't edit COB file names
end;

initialization
  RegisterEditor('PRAY Agent Files', 'agent', 'PRAY Editor', TPRAYDecompilerForm);
  RegisterEditor('PRAY Agent(s) Files', 'agents', 'PRAY Editor', TPRAYDecompilerForm);
  RegisterEditor('PRAY Family Files', 'family', 'PRAY Editor', TPRAYDecompilerForm);
  RegisterEditor('PRAY Creature Files', 'creature', 'PRAY Editor', TPRAYDecompilerForm);
  RegisterEditor('PRAY Blueprint Files', 'blueprint', 'PRAY Editor', TPRAYDecompilerForm);
  RegisterEditor('COB Files', 'cob', 'COB Editor', TPRAYDecompilerForm);
end.
