{
  TCreatureLabsCommLink, based on version 0.9 of TEvolutionEngineCommunicator
  Copyright (c) 2000 Michael Milburn (Fuzzie)
  Copyright (c) 1999-2000 Daniel Silverstone (Kinnison) of Creature Labs Ltd

  This unit requires Delphi 3 or above. However, you *might* be able to
  fiddle it into working with Delphi 2, it'll need work however.

  CreatureLabsCommLink supports communicating with Creatures 1, Creatures 2,
  and any products using the Creatures Evolution Engine (eg Creatures 3). I'll
  be adding support for any new Creature Labs products (if possible) as soon
  as I work out their communication mechanisms.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}

// How to use this component:

// Basically, set 'GameName' to the name of the game: 'Creatures 1',
// 'Creatures 2', or the CEE gamename, eg 'Creatures 3'.

// Call 'Connect' to establish a communication link, then set 'Transaction'
// to any string value to send that value as a command. 'Results' will then
// hold the return value. Finally, 'Disconnect' will drop the link.

unit CreatureLabsCommLink;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ActiveX, StdVCL, OleCtrls, ddeman;

const
  LIBID_Creatures2Ole: TGUID = '{22582D65-4EE0-11D1-8791-0060B07BFA18}';
  Class_Document: TGUID = '{22582D67-4EE0-11D1-8791-0060B07BFA18}';

type
  // The Creatures 2 OLE Automation interface
  ICreatures2SFC = interface
    ['{22582D66-4EE0-11D1-8791-0060B07BFA18}']
    function RequestMacro(var p_Variant1: OleVariant; out p_Variant2: OleVariant): WordBool; dispid 1;
    function ExecuteMacro(var p_Variant1, p_Variant2: OleVariant): WordBool; dispid 2;
    function CreateMacro(var p_Variant1, p_Variant2: OleVariant): WordBool; dispid 3;
    function DestroyMacro(var p_Variant1, p_Variant2: OleVariant): WordBool; dispid 4;
    function LoadMacro(var p_Variant1, p_Variant2: OleVariant): WordBool; dispid 5;
    function CreateCommand(Type_: Smallint): Integer; dispid 6;
    function DestroyCommand(Cmd: Integer): WordBool; dispid 7;
    procedure LoadCommand(Cmd: Integer; const String_: WideString); dispid 8;
    function RequestCommand(Cmd: Integer; var Data: WideString): WordBool; dispid 9;
    function FireCommand(Type_: Smallint; const Cmd: WideString; var Data: WideString): WordBool; dispid 10;
  end;

  // Creatures Evolution Engine (CEE/c2e) interface record
  PInterfaceRecord = ^TInterfaceRecord;
  TInterfaceRecord = packed record
    Magic: array [0..3] of Char; // should be 'c2e@'
    ServerID: Cardinal; // process id
    ReturnCode: Integer; // the return code of the last request
    DataSize: Cardinal; // the amount of data in the buffer
    BufferSize: Cardinal; // the total size of the buffer
    Pad: Integer; // padding so structure is aligned
  end;

  TCreatureLabsCommLink = class(TComponent)
  private
    // Creatures 3 definitions
    FMutex, FRequestEvent, FResultEvent, FMappedFile: Cardinal;
    FInterface: PInterfaceRecord; FAfterInterface: Pointer;

    // Creatures 2 definitions
    OLEInterface: ICreatures2SFC;

    // Shared definitions
    FConnected: boolean;
    FTransaction, FResults, FGameName: String;
    FReturnCode: Integer;

    dde: TDDEClientConv;
  protected
    procedure DoSetConnected(aValue: boolean);
    procedure DoSetGameName(aValue: string);
    procedure DoSetTransaction(aValue: string);
    procedure DoTransactionWithServer;
    procedure DoDDETransaction;
    procedure DoOLETransaction;

    function ConnectCEE: boolean;
    function ConnectDDE: boolean;
    function ConnectOLE: boolean;

    procedure DisconnectCEE;
    procedure DisconnectDDE;
    procedure DisconnectOLE;
  public
    scrip: Boolean; // C3 only: if set, we don't place the 'execute' header

    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    function Connect: boolean;
    procedure Disconnect;

  published
    property GameName: String read FGameName write DoSetGameName;
    property Connected: boolean read FConnected write DoSetConnected;
    property Transaction: String read FTransaction write DoSetTransaction;
    property Results: String read FResults;
    property ReturnCode: Integer read FReturnCode;
  end;

procedure Register;

implementation

uses ComObj;

procedure TCreatureLabsCommLink.DoDDETransaction;
begin
  if dde = nil then begin FResults := '***ERROR: dde is nil***'; end;

  dde.PokeData('Macro', PChar(FTransaction));
  FResults := dde.RequestData('Macro');
end;

procedure TCreatureLabsCommLink.DisconnectDDE;
begin
  dde.CloseLink;
  dde.Free;
  dde := nil;
end;

function TCreatureLabsCommLink.ConnectDDE: boolean;
begin
  dde := TDDEClientConv.Create(Self);
  dde.ConnectMode := ddeManual; 
  dde.SetLink('Vivarium', 'Macro');
  Result := dde.OpenLink;
end;

procedure TCreatureLabsCommLink.DoSetTransaction(aValue : string);
begin
  FTransaction := aValue;
  if FConnected then begin
    if FGameName = 'Creatures 1' then DoDDETransaction
    else if FGameName = 'Creatures 2' then DoOLETransaction
    else DoTransactionWithServer;
  end else begin
    FResults := '***ERROR: not connected***';
    Exit;
  end;
end;

procedure TCreatureLabsCommLink.DoOLETransaction;
var
  r: WideString;
begin
  OLEInterface.FireCommand(1, FTransaction, r);
  FResults := r;
end;

procedure TCreatureLabsCommLink.DoTransactionWithServer;
var
  SerPID : Cardinal;
  TwoHands : array[0..1] of Cardinal;
  tempString : string;
begin
  FResults := '***ERROR***';
  if scrip then tempString := FTransaction + #0
  else tempString := 'execute' + #10 + FTransaction + #0;

  if (Cardinal(Length(tempString)) > FInterface^.BufferSize) then begin
    FResults := '***ERROR: buffer too small***';
    exit;
  end else if (FInterface^.Magic <> 'c2e@') then begin
    FResults := '***ERROR: majic not valid (corrupted buffer?)***';
    exit;
  end else if (WaitForSingleObject(FMutex, 500) <> WAIT_OBJECT_0) then begin
    FResults := '***ERROR: couldn''t obtain mutex***';
    exit;
  end;
  CopyMemory(FAfterInterface, PChar(tempString), Length(tempString));
  FInterface^.DataSize := Length(tempString);
  ResetEvent(FResultEvent);
  SetEvent(FRequestEvent);
  SerPID := OpenProcess(PROCESS_ALL_ACCESS, false, FInterface^.ServerID);
  TwoHands[0] := FResultEvent;
  TwoHands[1] := SerPID;
  if (WaitForMultipleObjects(2, @(TwoHands[0]), false, INFINITE) <> WAIT_OBJECT_0) then
  begin
    CloseHandle(SerPID);
    ReleaseMutex(FMutex);
    FResults := '***ERROR: couldn''t wait for objects***';
    exit;
  end;
  CloseHandle(SerPID);
  SetLength(FResults, FInterface^.DataSize - 1);
  CopyMemory(PChar(FResults), FAfterInterface, FInterface^.DataSize - 1);
  FReturnCode := FInterface^.ReturnCode;
  ReleaseMutex(FMutex);
  FTransaction := #0;
  CopyMemory(FAfterInterface, PChar(FTransaction), 1);
  FInterface^.DataSize := 1;
end;

procedure TCreatureLabsCommLink.DoSetGameName(aValue : string);
var
 was : boolean;
begin
  was := FConnected;
  if (FConnected) then
    Disconnect;
  FGameName := aValue;
  if (was) then
    Connect;
end;

procedure TCreatureLabsCommLink.DoSetConnected(aValue : boolean);
begin
  if (aValue = false) then
    Disconnect
  else
    Connect;
end;

constructor TCreatureLabsCommLink.Create(aOwner : TComponent);
begin
  inherited Create(aOwner);
  FMutex := 0;
  FRequestEvent := 0;
  FResultEvent := 0;
  FMappedFile := 0;
  FInterface := nil;
  FConnected := false;
  FGameName := '';
  FTransaction := '';
  FResults := '***ERROR***';
  scrip := False;
  dde := nil;
end;

destructor TCreatureLabsCommLink.Destroy;
begin
  if (FConnected) then
    Disconnect;
  inherited Destroy;
end;

function TCreatureLabsCommLink.Connect: Boolean;
begin
  if (FConnected) then
    Disconnect;

  if FGameName = 'Creatures 1' then Result := ConnectDDE
  else if FGameName = 'Creatures 2' then Result := ConnectOLE
  else Result := ConnectCEE;

  FConnected := Result;
end;

function TCreatureLabsCommLink.ConnectOLE: Boolean;
begin
  Result := True;

  try
    OLEInterface := ICreatures2SFC(CreateOLEObject('SFC2.OLE'));
  except
    on Exception do Result := False;
  end;
end;

function TCreatureLabsCommLink.ConnectCEE: Boolean;
var
  tempString : String;
begin
  Result := false;
  // ... Do connection stuffs
  tempString := FGameName + '_mutex';
  FMutex := OpenMutex( MUTEX_ALL_ACCESS, false, PChar(tempString) );
  if FMutex = 0 then begin
    Disconnect;
    exit;
  end;
  tempString := FGameName+'_request';
  FRequestEvent := OpenEvent( EVENT_ALL_ACCESS, false, PChar(tempString) );
  if FRequestEvent = 0 then begin
    Disconnect;
    exit;
  end;
  tempString := FGameName+'_result';
  FResultEvent := OpenEvent( EVENT_ALL_ACCESS, false, PChar(tempString) );
  if FResultEvent = 0 then begin
    Disconnect;
    exit;
  end;
  tempString := FGameName+ '_mem';
  FMappedFile := OpenFileMapping(FILE_MAP_ALL_ACCESS, false, PChar(tempString) );
  if FMappedFile = 0 then begin
    Disconnect;
    exit;
  end;
  FInterface := PInterfaceRecord(MapViewOfFile(FMappedFile,FILE_MAP_ALL_ACCESS,0,0,0));
  if FInterface = nil then begin
    Disconnect;
    exit;
  end;
  FAfterInterface := Pointer(Cardinal(FInterface) + sizeof(TInterfaceRecord));
  // Sanity Check on the header
  if FInterface^.Magic <> 'c2e@' then begin
    Disconnect;
    exit;
  end;
  if (FTransaction <> '') then
    DoTransactionWithServer;
  Result := true;
end;

procedure TCreatureLabsCommLink.Disconnect;
begin
  if FGameName = 'Creatures 1' then DisconnectDDE
  else if FGameName = 'Creatures 2' then DisconnectOLE
  else DisconnectCEE;

  FConnected := false;
end;

procedure TCreatureLabsCommLink.DisconnectOLE;
begin
  // Umm, you should release OLEInterface here.
  // However, I can't actually work out *how* ... Delphi normally automatically releases OLE automation objects once they go out of scope, and it doesn't want to let me release it manually.
  // Evil COM interfaces. Grr.
end;

procedure TCreatureLabsCommLink.DisconnectCEE;
begin
  if not FConnected = false then Exit;

  // ... Do disconnect Stuffs
  if (FInterface <> nil) then
  begin
    UnmapViewOfFile(FInterface);
    FInterface := nil;
  end;
  if (FMappedFile <> 0) then
  begin
    CloseHandle(FMappedFile);
    FMappedFile := 0;
  end;
  if (FResultEvent <> 0) then
  begin
    CloseHandle(FResultEvent);
    FResultEvent := 0;
  end;
  if (FRequestEvent <> 0) then
  begin
    CloseHandle(FRequestEvent);
    FRequestEvent := 0;
  end;
  if (FMutex <> 0) then
  begin
    CloseHandle(FMutex);
    FMutex := 0;
  end;
end;

procedure Register;
begin
  RegisterComponents('C2e', [TCreatureLabsCommLink]);
end;

end.
