unit CAOSSyntaxMemo;

// A heavily modified copy of TSyntaxMemo from Quark 5. (http://www.planetquake.com/quark/)
// Partially Copyright (c) Armin Rigo

// (fixme: this needs a rewrite, (a) move to TRichEdit - umm ... the cursor doesn't redraw and the view is messed up ... UNTIL you select off the bottom of the view ... then it works fine
//                               (b) rewrite GDI code to use Delphi-style stuff?)

// Also: it doesn't work properly under WINE (UNIX Windows emulator) ... the
// cursor is out of sync, maybe the fonts system is wierd? look into it...

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls;

type
  TFormatLineInfo  = record
    Start, Length: Integer;
    Font: TFont;
  end;
  TFormatLineArray = array[0..127] of TFormatLineInfo;
  TFormatLineEvent = procedure(I: Integer; const S: String; var F: TFormatLineArray) of object;

{$IFDEF RICHEDIT}
  TSyntaxMemo = class(TRichEdit)
{$ELSE}
  TSyntaxMemo = class(TMemo)
{$ENDIF}
  private
    FSelection: Boolean;
    FTabStops: Integer;
    FOnFormatLine: TFormatLineEvent;
    FOldFontChange: TNotifyEvent;
    procedure wmPaint(var Msg: TMessage); message wm_Paint;
    procedure wmCommand(var Msg: TMessage); message wm_Command;
    procedure SetTabStops(T: Integer);
    procedure FontChange(Sender: TObject);
  protected
    procedure Change; override;
    procedure PaintWindow1(DC: HDC);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SelectRange(Start, Length: Integer);
  published
    property OnFormatLine: TFormatLineEvent read FOnFormatLine write FOnFormatLine;
    property TabStops: Integer read FTabStops write SetTabStops default 4;
    property WantTabs default True;
    property ScrollBars default ssBoth;
    property WordWrap default False;
    property Align;
    property Font;
    property OnChange;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Additional', [TSyntaxMemo]);
end;

constructor TSyntaxMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Parent := AOwner as TWinControl;
  WantTabs := True;
  WordWrap := False;
  ScrollBars := ssBoth;
  FTabStops := 4;
  FOldFontChange := Font.OnChange;
  Font.OnChange := FontChange;
end;

procedure TSyntaxMemo.SetTabStops;
begin
  if T <= 0 then Exit;
  FTabStops := T;
  if HandleAllocated then begin
    T := T * 4;
    SendMessage(Handle, em_SetTabStops, 1, LongInt(@T));
    Invalidate;
  end;
end;

procedure TSyntaxMemo.FontChange;
begin
  TabStops := TabStops;
  if Assigned(FOldFontChange) then
    FOldFontChange(Sender);
end;

procedure TSyntaxMemo.CreateWnd;
begin
  inherited CreateWnd;
  TabStops := TabStops;
end;

procedure TSyntaxMemo.wmPaint;
var
  DC: HDC;
  PaintInfo: TPaintStruct;
begin
  if (csDesigning in ComponentState) or (SelLength > 0) then inherited else begin
    DC := BeginPaint(Handle, PaintInfo);
    try
      PaintWindow1(DC);
    finally
      EndPaint(Handle, PaintInfo);
    end;
  end;
end;

procedure TSyntaxMemo.Change;
begin
  inherited;
  InvalidateRect(Handle, Nil, False);
end;

procedure TSyntaxMemo.PaintWindow1;
var
  Taille: TSize;
  I, J, DispPos, DispRectLeft: Integer;
  FormatRect, DisplayRect: TRect;
  S, TabSpaces: String;
  Font0, Font1: HFont;
  FormatFont: TFormatLineArray;
  FormatInfo: ^TFormatLineInfo;
  Color1: TColorRef;
  Metrics: TTextMetric;

  // write a line to the screen...
  procedure Write(L: Integer; l_Font: HFont; l_Color: TColorRef);
  var
    I: Integer;
    Tabs: String;
    x: HGDIObj;
  begin
    if (J + L) > Length(S) then
      L := Length(S) - J;
    if L > 0 then begin
      x := SelectObject(DC, l_Font); // select font
      SetTextColor(DC, l_Color); // select colorref
      Tabs := Copy(S, J + 1, L);
      I := 1;
      while I <= Length(Tabs) do begin
        if Tabs[I] = chr(vk_Tab) then Tabs := Copy(Tabs, 1, I - 1) + Copy(TabSpaces, (DispPos + I) mod FTabStops + 1, MaxInt) + Copy(Tabs, I + 1, MaxInt);
        Inc(I);
      end;
      Inc(DispPos, Length(Tabs));
      FormatRect.Right := FormatRect.Left + Length(Tabs) * Taille.cx;
      ExtTextOut(DC, FormatRect.Left, FormatRect.Top + Metrics.tmAscent, eto_Opaque, @FormatRect, PChar(Tabs), Length(Tabs), Nil);
      FormatRect.Left := FormatRect.Right;
      Inc(J, L);
      SelectObject(DC, x); // unselect font
    end;
  end;

begin
  // setup a string to replace tabs
  SetLength(TabSpaces, TabStops);
  FillChar(TabSpaces[1], TabStops, ord(' '));

  // get/setup the display rectangle
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@DisplayRect));
  DispRectLeft := DisplayRect.Left;
  I := GetScrollPos(Handle, SB_HORZ);
  Dec(DisplayRect.Left, I);

  // get the first visible line
  I := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0,0);

  // setup GDI stuff
  Font1 := Font.Handle;
  Color1 := ColorToRGB(Font.Color);
  Font0 := SelectObject(DC, Font1); // select default font in

  try
    GetTextExtentPoint32(DC, 'X', 1, Taille);
    SetTextAlign(DC, ta_Left or ta_BaseLine);
    GetTextMetrics(DC, Metrics);
    FormatRect := DisplayRect;
    while FormatRect.Top < DisplayRect.Bottom do begin
      // fill in dummy chars
      FillChar(FormatFont, SizeOf(FormatFont), $3F);

      if I < Lines.Count then begin
        S := Lines[I];
        if Assigned(FOnFormatLine) then
          FOnFormatLine(I, S, FormatFont);
      end else
        S := '';
      // sort out the rects
      FormatRect.Bottom := FormatRect.Top + Taille.cy;
      FormatRect.Left := DisplayRect.Left;
      J := 0;
      DispPos := -1;
      // point to the first format information
      FormatInfo := @FormatFont[0];
      while J < Length(S) do
        with FormatInfo^ do begin
          Write(Start-J, Font1, Color1);
          if Start = $3F3F3F3F then Break; // if OnFormatLine hasn't got this far, proceed to next line
          Write(Length, Font.Handle, ColorToRGB(Font.Color));
          // Inc(FormatInfo);
          // point to the next format information
          FormatInfo := Pointer(Integer(FormatInfo) + SizeOf(TFormatLineInfo));
        end;
      // blt it :)
      PatBlt(DC, FormatRect.Left, FormatRect.Top, DisplayRect.Right - FormatRect.Left, Taille.cy, Whiteness);
      // next line
      Inc(I);
      // and update the rects for the next line...
      FormatRect.Top := FormatRect.Bottom;
    end;
  finally
    SelectObject(DC, Font0); // swap default font out
  end;
  PatBlt(DC, 0, 0, DispRectLeft, DisplayRect.Bottom, Whiteness);
end;

procedure TSyntaxMemo.KeyDown;
begin
 inherited;
 PostMessage(Handle, wm_Command, 0, 0);
end;

procedure TSyntaxMemo.MouseDown;
begin
 inherited;
 PostMessage(Handle, wm_Command, 0, 0);
end;

procedure TSyntaxMemo.MouseUp;
begin
 inherited;
 PostMessage(Handle, wm_Command, 0, 0);
end;

procedure TSyntaxMemo.wmCommand;
begin
 if FSelection xor (SelLength>0) then
  begin
{   if FSelection then
    Paint;}
     InvalidateRect(Handle, Nil, False);
   FSelection:=not FSelection;
  end;
end;

procedure TSyntaxMemo.SelectRange(Start, Length: Integer);
begin
 SendMessage(Handle, em_SetSel, Start, Start);
 SendMessage(Handle, em_ScrollCaret, 0,0);
 Update;
 SendMessage(Handle, em_SetSel, Start, Start+Length);
end;

end.
