program zeus;

{

    Zeus, a development tool for the Creatures series
    Copyright (c) 1998-2000 Michael 'Fuzzie' Milburn

    contact e-mail: fuzzie@warpedgames.com
    http://www.warpedgames.com/

    Zeus is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Zeus is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}

{%ToDo 'zeus.todo'}

uses
  Forms,
  SysUtils,
  Registry,
  Windows,
  CreatureLabsCommLink,
  Unit1 in 'shared\Unit1.pas' {ZeusMainForm},
  unitAboutBox in 'shared\unitAboutBox.pas' {AboutBox},
  unitChildForm in 'shared\unitChildForm.pas' {MDIChildForm},
  AgentClasses in 'DelphiCore\AgentClasses.pas',
  BLKClasses in 'DelphiCore\blkclasses.pas',
  RoomStuff in 'DelphiCore\RoomStuff.pas',
  ZeusClasses in 'DelphiCore\ZeusClasses.pas',
  unitPRAYDecompiler in 'prayeditor\unitPRAYDecompiler.pas' {PRAYDecompilerForm},
  unitCatalogueForm in 'catalogueeditor\unitCatalogueForm.pas' {CatalogueEditorForm},
  unitMapViewer in 'roomeditor\unitMapViewer.pas' {MapViewerForm},
  unitMapEditor in 'roomeditor\unitMapEditor.pas' {MetaroomEditorForm},
  unitBLKCreator in 'blkeditor\unitBLKCreator.pas' {BLKCreatorForm},
  unitBLKViewer in 'blkeditor\unitBLKViewer.pas' {BLKViewerForm},
  unitCAOSReference in 'agenteditor\unitCAOSReference.pas' {CAOSRefForm},
  unitCAOSCalculator in 'agenteditor\unitCAOSCalculator.pas' {CAOSCalculatorForm},
  unitNewAgentWizard in 'agenteditor\unitNewAgentWizard.pas' {NewAgentWizardForm},
  unitPRAYWizard in 'agenteditor\unitPRAYWizard.pas' {PRAYCompileForm},
  unitAgentForm in 'agenteditor\unitAgentForm.pas' {AgentForm},
  unitWelcome in 'shared\unitWelcome.pas' {WelcomeForm},
  unitSyntaxInformation in 'tokenstuff\unitSyntaxInformation.pas',
  unitTokeniser in 'tokenstuff\unitTokeniser.pas',
  unitSyntaxCheck in 'tokenstuff\unitSyntaxCheck.pas',
  unitSplahScreen in 'shared\unitSplahScreen.pas' {SplashForm},
  unitNewFileForm in 'shared\unitNewFileForm.pas' {NewFileForm},
  unitCommLinkForm in 'shared\unitCommLinkForm.pas' {CommLinkForm},
  unitErrorHandler in 'shared\unitErrorHandler.pas';

{$R *.RES}

// if you do
//   zeus a.cos b.cos "blah c.cos"
// then it'll open those three files
procedure zeus_init_app_params;
var
  zeusxi: Integer;
begin
  zeusxi := 0;
  while zeusxi < ParamCount do begin // <-- 'for' doesn't want to work
    Inc(zeusxi);
    ZeusMainForm.OpenFile(ParamStr(zeusxi));
  end;
end;

var
  rr: TRegistry;
begin
  // Application.Initialize; is called in initialization clause

  // various CreateForm calls are in unit initialization clauses
  
  Application.CreateForm(TCAOSRefForm, CAOSRefForm);
  Application.CreateForm(TCAOSCalculatorForm, CAOSCalculatorForm);
  Application.CreateForm(TCommLinkForm, CommLinkForm);
  SplashForm.StatusLabel.Caption := 'Loading Settings...'; SplashForm.Refresh;

  rr := TRegistry.Create;
  if rr.OpenKey('Software\Creature Labs\Zeus', False) then begin
    try
      CommLink.GameName := rr.ReadString('GameName'); WelcomeForm.CheckBox1.Checked := rr.ReadBool('ShowWelcome'); ColorCode := rr.ReadBool('ColorCode'); SyntaxCheck := rr.ReadBool('SyntaxCheck'); CAOSRefSync := rr.ReadBool('CAOSRefSync'); autoindent := rr.ReadBool('AutoIndent');
      recentfiles.Add(rr.ReadString('RecentFile1')); recentfiles.Add(rr.ReadString('RecentFile2')); recentfiles.Add(rr.ReadString('RecentFile3')); recentfiles.Add(rr.ReadString('RecentFile4'));
      recentfiles.Add(rr.ReadString('RecentFile5')); recentfiles.Add(rr.ReadString('RecentFile6')); recentfiles.Add(rr.ReadString('RecentFile7')); recentfiles.Add(rr.ReadString('RecentFile8'));
    except
      on ERegistryException do Application.MessageBox('You have upgraded your version of Zeus. New settings may not be setup correctly.', 'Zeus', MB_OK + MB_ICONWARNING);
    end;
  end else begin
    Application.MessageBox('Welcome to Zeus! Zeus is currently set to Creatures 3 mode. You''ll need to change this if you''re using another game.', 'Zeus', MB_OK + MB_ICONINFORMATION);

    // set some defaults
    CommLink.GameName := 'Creatures 3'; WelcomeForm.CheckBox1.Checked := True;
    ColorCode := True; SyntaxCheck := False; CAOSRefSync := True; AutoIndent := False; // buggy
  end;
  rr.Free;

  while recentfiles.Count < 8 do recentfiles.Add(''); // make sure the recent files buffer is big enough

  SplashForm.StatusLabel.Caption := 'Reading Game Content...'; SplashForm.Refresh;
  SyntaxDataUpdate;
  SplashForm.StatusLabel.Caption := 'Finished!'; SplashForm.Refresh; Sleep(500);
  SplashForm.Free;

  ZeusMainForm.Show;

  if ParamCount > 0 then zeus_init_app_params
  else if WelcomeForm.CheckBox1.Checked then WelcomeForm.ShowModal;

  Application.Run;

  rr := TRegistry.Create;
  if rr.OpenKey('Software\Creature Labs\Zeus', True) then begin
    try
      rr.WriteString('GameName', CommLink.GameName);
      rr.WriteBool('ShowWelcome', WelcomeForm.CheckBox1.Checked);
    except
      on Exception do Application.MessageBox('When Zeus was trying to write settings to the registry, the key opened fine but an exception occurred when trying to write. Null pointers? Contact Fuzzie...', 'Zeus', MB_OK + MB_ICONERROR);
    end;

    rr.WriteBool('ColorCode', ColorCode); rr.WriteBool('SyntaxCheck', SyntaxCheck); rr.WriteBool('CAOSRefSync', CAOSRefSync); rr.WriteBool('AutoIndent', autoindent);
    rr.WriteString('RecentFile1', recentfiles[0]); rr.WriteString('RecentFile2', recentfiles[1]); rr.WriteString('RecentFile3', recentfiles[2]); rr.WriteString('RecentFile4', recentfiles[3]);
    rr.WriteString('RecentFile5', recentfiles[4]); rr.WriteString('RecentFile6', recentfiles[5]); rr.WriteString('RecentFile7', recentfiles[6]); rr.WriteString('RecentFile8', recentfiles[7]);
  end else begin
    Application.MessageBox('Zeus couldn''t save your settings to the registry. Any changes to Zeus settings made will be lost.', 'Zeus', MB_OK + MB_ICONERROR);
  end;
  rr.Free;
end.
